"""File and emoji objects for the Notion API."""

from __future__ import annotations

from typing import cast
from urllib.parse import urlparse

from emoji import emojize, is_emoji

from ultimate_notion.core import Wrapper, get_repr
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.rich_text import Text, html_img

NOTION_HOSTED_DOMAIN = 'secure.notion-static.com'


class FileInfo(Wrapper[objs.FileObject], wraps=objs.FileObject):
    """Information about a web resource e.g. for the files property."""

    obj_ref: objs.FileObject

    def __init__(self, *, url: str, name: str | None = None, caption: str | None = None) -> None:
        caption_obj = Text(caption).obj_ref if caption is not None else None  # [] is not accepted here by the API!
        if is_notion_hosted(url):
            self.obj_ref = objs.HostedFile.build(url=url, name=name, caption=caption_obj)
        else:
            self.obj_ref = objs.ExternalFile.build(url=url, name=name, caption=caption_obj)

    @classmethod
    def wrap_obj_ref(cls, obj_ref: objs.FileObject) -> FileInfo:
        self = cast(FileInfo, cls.__new__(cls))
        self.obj_ref = obj_ref
        return self

    def __eq__(self, other: object) -> bool:
        match other:
            case str():
                return str(self) == other
            case FileInfo():
                return str(self) == str(other)
            case _:
                return NotImplemented

    def __hash__(self):
        return hash(str(self))

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.url

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this file."""
        return html_img(self.url, size=2)

    @property
    def name(self) -> str | None:
        return self.obj_ref.name

    @name.setter
    def name(self, name: str | None) -> None:
        self.obj_ref.name = name

    @property
    def caption(self) -> Text:
        return Text.wrap_obj_ref(self.obj_ref.caption)

    @caption.setter
    def caption(self, caption: str | None) -> None:
        self.obj_ref.caption = Text(caption).obj_ref if caption is not None else []

    @property
    def url(self) -> str:
        return self.obj_ref.value.url

    @url.setter
    def url(self, url: str) -> None:
        self.obj_ref.value.url = url


class Emoji(Wrapper[objs.EmojiObject], str, wraps=objs.EmojiObject):
    """Emoji object which behaves like str."""

    def __init__(self, emoji: str) -> None:
        if not is_emoji(emoji):
            emoji = emojize(emoji)
        if not is_emoji(emoji):
            msg = f'Invalid emoji string: {emoji}'
            raise ValueError(msg)
        self.obj_ref = objs.EmojiObject.build(emoji)

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.obj_ref.emoji

    def __eq__(self, other: object) -> bool:
        match other:
            case str():
                return str(self) == other
            case Emoji():
                return str(self) == str(other)
            case _:
                return NotImplemented

    def __hash__(self):
        return hash(str(self))

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this file."""
        return str(self)

    def to_code(self) -> str:
        """Represent the emoji as :shortcode:, e.g. :smile:"""
        raise NotImplementedError

    @classmethod
    def from_code(cls, shortcode: str) -> Emoji:
        """Create an Emoji object from a :shortcode:, e.g. :smile:"""
        raise NotImplementedError


def wrap_icon(icon_obj: objs.FileObject | objs.EmojiObject) -> FileInfo | Emoji:
    """Wrap the icon object into the corresponding class."""
    match icon_obj:
        case objs.ExternalFile():
            return FileInfo.wrap_obj_ref(icon_obj)
        case objs.EmojiObject():
            return Emoji.wrap_obj_ref(icon_obj)
        case _:
            msg = f'unknown icon object of {type(icon_obj)}'
            raise RuntimeError(msg)


def is_notion_hosted(url: str) -> bool:
    """Check if the URL is hosted on Notion."""
    return urlparse(url).netloc == NOTION_HOSTED_DOMAIN



"""Blocks that make up the content of a page."""

from __future__ import annotations

import itertools
import mimetypes
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, TypeAlias, TypeGuard, TypeVar, cast, overload

from tabulate import tabulate
from typing_extensions import Self

from ultimate_notion.comment import Comment, Discussion
from ultimate_notion.core import InvalidAPIUsageError, NotionEntity, get_active_session, get_url
from ultimate_notion.file import Emoji, FileInfo, wrap_icon
from ultimate_notion.markdown import md_comment
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.enums import BGColor, CodeLang, Color
from ultimate_notion.rich_text import Text, User
from ultimate_notion.utils import del_nested_attr

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.page import Page


DO = TypeVar('DO', bound=obj_blocks.DataObject)  # ToDo: Use new syntax when requires-python >= 3.12


class DataObject(NotionEntity[DO], wraps=obj_blocks.DataObject):
    """The base type for all data-related types, i.e, pages, databases and blocks."""

    @property
    def block_url(self) -> str:
        """Return the URL of the block.

        !!! note

            Databases and pages are also considered blocks in Notion but they also have a
            long-form URL. Use the `url` property to get the long-form URL.
        """
        return get_url(self.id)

    @property
    def last_edited_by(self) -> User:
        """Return the user who last edited the block."""
        session = get_active_session()
        return session.get_user(self.obj_ref.last_edited_by.id)

    @property
    def has_children(self) -> bool:
        """Return whether the object has children."""
        return self.obj_ref.has_children

    def _delete_me_from_parent(self) -> None:
        """Remove the block from the parent's children list."""
        if isinstance(self.parent, ChildrenMixin) and self.parent._children is not None:
            for idx, child in enumerate(self.parent._children):
                if child.id == self.id:
                    del self.parent._children[idx]
                    break
            self.parent.obj_ref.has_children = bool(self.parent._children)

    def delete(self) -> Self:
        """Delete the block.

        Pages and databases are moved to the trash, blocks are deleted permanently.
        """
        session = get_active_session()
        self.obj_ref = cast(DO, session.api.blocks.delete(self.id))
        self._delete_me_from_parent()
        return self

    @property
    def is_deleted(self) -> bool:
        """Return wether the object is in trash."""
        return self.obj_ref.in_trash or self.obj_ref.archived

    @abstractmethod
    def to_markdown(self) -> str:
        """Return the content of the block as Markdown."""
        ...

    def _to_markdown(self) -> str:
        """Return the content of the block as Markdown.

        This method can be overridden to provide a custom Markdown representation,
        e.g. for a page/database child block.
        """
        return self.to_markdown()


class ChildrenMixin(DataObject[DO], wraps=obj_blocks.DataObject):
    """Mixin for data objects that can have children

    Note that we don't use the `children` property of some Notion objects, e.g. paragraph, quote, etc.,
    as not every object has this property, e.g. a page, database or toggable heading.
    """

    # ToDo: This could be reworked to differentiate between blocks that have a children attribute and those that don't,
    # like headings and pages. In the former case we would not neet to have a _children attribute and could directly
    # access the children attribute of the object reference. In the latter case we would need to have a _children.
    # This would save some memory and appending would work for the former case even for blocks that are not
    # yet in Notion.

    _children: list[Block] | None = None

    def _gen_children_cache(self) -> list[Block]:
        """Generate the children cache."""
        if self.is_deleted:
            msg = 'Cannot retrieve children of a deleted block from Notion.'
            raise RuntimeError(msg)

        session = get_active_session()
        child_blocks_objs = list(session.api.blocks.children.list(parent=objs.get_uuid(self.obj_ref)))
        self.obj_ref.has_children = bool(child_blocks_objs)  # update the property manually to avoid API call
        if isinstance(self.obj_ref, obj_blocks.Block) and hasattr(self.obj_ref.value, 'children'):
            self.obj_ref.value.children = child_blocks_objs  # update the children attribute

        child_blocks: list[Block] = [Block.wrap_obj_ref(block) for block in child_blocks_objs]

        for idx, child_block in enumerate(child_blocks):
            if isinstance(child_block, ChildPage):
                child_blocks[idx] = cast(Block, child_block.page)
            elif isinstance(child_block, ChildDatabase):
                child_blocks[idx] = cast(Block, child_block.db)

        return [cast(Block, session.cache.setdefault(block.id, block)) for block in child_blocks]

    @property
    def children(self) -> tuple[Block, ...]:
        """Return the children of this block."""
        if self.in_notion:
            if self._children is None:
                self._children = self._gen_children_cache()
            children = tuple(self._children)  # we copy implicitly to allow deleting blocks while iterating over it
        elif isinstance(self.obj_ref, obj_blocks.Block) and hasattr(self.obj_ref.value, 'children'):
            children = tuple(Block.wrap_obj_ref(block) for block in self.obj_ref.value.children)
        else:
            children = ()
        return children

    def append(self, blocks: Block | Sequence[Block], *, after: Block | None = None) -> Self:
        """Append a block or a sequence of blocks to the content of this block."""
        if not self.in_notion:
            msg = 'Cannot append blocks to a block that is not in Notion.'
            raise RuntimeError(msg)

        blocks = [blocks] if isinstance(blocks, Block) else blocks
        if not blocks:
            return self
        for block in blocks:
            if not isinstance(block, Block):
                msg = f'Cannot append {type(block)} to a block.'
                raise ValueError(msg)
            if block.in_notion:
                msg = 'Cannot append a block that is already in Notion.'
                raise InvalidAPIUsageError(msg)

        if self._children is None:
            self._children = self._gen_children_cache()

        session = get_active_session()
        block_objs = [block.obj_ref for block in blocks]
        after_obj = None if after is None else after.obj_ref
        block_objs, after_block_objs = session.api.blocks.children.append(self.obj_ref, block_objs, after=after_obj)

        for block, obj in zip(blocks, block_objs, strict=True):
            block.obj_ref = obj
            session.cache[block.id] = block

        if after is None:
            self._children.extend(blocks)
        else:
            insert_idx = next(idx for idx, block in enumerate(self._children) if block.id == after.id) + 1
            # we update the blocks after the position we want to insert.
            for block, updated_block_obj in zip(self._children[insert_idx:], after_block_objs, strict=True):
                block.obj_ref.update(**updated_block_obj.model_dump())
            self._children[insert_idx:insert_idx] = blocks

        self.obj_ref.has_children = True
        return self


class CommentMixin(DataObject[DO], wraps=obj_blocks.DataObject):
    """Mixin for objects that can have comments and discussions."""

    _comments: list[Discussion] | None = None

    @staticmethod
    def _group_by_discussions(comments: list[Comment]) -> list[list[Comment]]:
        comments.sort(key=lambda comment: comment.discussion_id)
        grouped = [list(group) for _, group in itertools.groupby(comments, key=lambda comment: comment.discussion_id)]
        grouped.sort(key=lambda group: group[0].created_time)
        return grouped

    def _generate_comments_cache(self) -> list[Discussion]:
        """Generate the comments cache."""
        if self.is_deleted:
            msg = 'Cannot retrieve comments of a deleted block from Notion.'
            raise RuntimeError(msg)

        session = get_active_session()
        comment_objs = session.api.comments.list(self.obj_ref.id)
        comments = [Comment.wrap_obj_ref(comment_obj) for comment_obj in comment_objs]
        return [Discussion(comment_thread, parent=self) for comment_thread in self._group_by_discussions(comments)]

    @property
    def _discussions(self) -> tuple[Discussion, ...]:
        """Return comments of this block or page as list of discussions, i.e. threads of comments."""
        if self.in_notion:
            if self._comments is None:
                self._comments = self._generate_comments_cache()
            comments = tuple(self._comments)
        else:
            comments = ()
        return comments


BT = TypeVar('BT', bound=obj_blocks.Block)  # ToDo: Use new syntax when requires-python >= 3.12


class Block(CommentMixin, DataObject[BT], ABC, wraps=obj_blocks.Block):
    """Abstract Notion block.

    Parent class of all block types.
    """

    @property
    def discussions(self) -> tuple[Discussion, ...]:
        """Return comments of this block as list of discussions, i.e. threads of comments."""
        return self._discussions

    def reload(self) -> Self:
        """Reload the block from the API."""
        session = get_active_session()
        self.obj_ref = cast(BT, session.api.blocks.retrieve(self.id))
        return self

    def _update_in_notion(self, *, exclude_attrs: Sequence[str] | None = None) -> None:
        """Update the locally modified block on Notion."""
        if self.in_notion:
            session = get_active_session()
            # missing_ok=True to cover `Heading` which behave like `TextBlock` but have no `children` attribute
            obj_ref = del_nested_attr(self.obj_ref, exclude_attrs, inplace=False, missing_ok=True)
            self.obj_ref = cast(BT, session.api.blocks.update(obj_ref))

    def replace(self, blocks: Block | Sequence[Block]) -> None:
        """Replace this block with another block or blocks."""
        if not isinstance(blocks, Sequence):
            blocks = [blocks]

        if self.is_deleted:
            msg = 'Cannot replace a deleted block.'
            raise InvalidAPIUsageError(msg)

        for block in blocks:  # do complete sanity check first
            if block.in_notion:
                msg = f'Cannot replace with a block {block} that is already in Notion.'
                raise InvalidAPIUsageError(msg)

        if self.parent is not None and isinstance(self.parent, ChildrenMixin):
            for block in reversed(blocks):
                self.parent.append(block, after=self)
        else:
            msg = 'Cannot replace a block that has no parent.'
            raise InvalidAPIUsageError(msg)

        self.delete()

    def insert_after(self, blocks: Block | Sequence[Block]) -> None:
        """Insert a block or several blocks after this block."""
        if not isinstance(blocks, Sequence):
            blocks = [blocks]

        if self.is_deleted:
            msg = 'Cannot insert a block after a deleted block.'
            raise InvalidAPIUsageError(msg)

        for block in blocks:  # do complete sanity check first
            if block.in_notion:
                msg = f'Cannot insert block {block} that is already in Notion.'
                raise InvalidAPIUsageError(msg)

        if self.parent is not None and isinstance(self.parent, ChildrenMixin):
            for block in reversed(blocks):
                self.parent.append(block, after=self)
        else:
            msg = 'Cannot insert a block that has no parent.'
            raise InvalidAPIUsageError(msg)


AnyBlock: TypeAlias = Block[Any]
"""For type hinting purposes, especially for lists of blocks, i.e. list[AnyBlock], in user code."""


class TextBlock(Block[BT], ABC, wraps=obj_blocks.TextBlock):
    """Abstract Text block.

    Parent class of all text block types.
    """

    def __init__(self, text: str) -> None:
        super().__init__()
        self.obj_ref.value.rich_text = Text(text).obj_ref

    @property
    def rich_text(self) -> Text:
        """Return the text content of this text block."""
        rich_texts = self.obj_ref.value.rich_text
        return Text.wrap_obj_ref(rich_texts)

    @rich_text.setter
    def rich_text(self, text: str) -> None:
        # Right now this leads to a mypy error, see
        # https://github.com/python/mypy/issues/3004
        # Always pass `RichText` objects to this setter to avoid this error.
        self.obj_ref.value.rich_text = Text(text).obj_ref
        self._update_in_notion(exclude_attrs=['paragraph.children'])


class Code(TextBlock[obj_blocks.Code], wraps=obj_blocks.Code):
    """Code block."""

    def __init__(
        self,
        text: str,
        *,
        language: CodeLang = CodeLang.PLAIN_TEXT,
        caption: str | None = None,
    ) -> None:
        super().__init__(text)
        self.obj_ref.value.language = language
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else []

    @property
    def language(self) -> CodeLang:
        return self.obj_ref.value.language

    @language.setter
    def language(self, language: CodeLang) -> None:
        self.obj_ref.value.language = language
        self._update_in_notion()

    @property
    def caption(self) -> Text:
        return Text.wrap_obj_ref(self.obj_ref.value.caption)

    @caption.setter
    def caption(self, caption: str | None) -> None:
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else []
        self._update_in_notion()

    def to_markdown(self) -> str:
        lang = self.obj_ref.value.language
        return f'```{lang}\n{self.rich_text.to_markdown()}\n```'


class ColoredTextBlock(TextBlock[BT], ABC, wraps=obj_blocks.TextBlock):
    """Abstract Text block with color.

    Parent class of all text block types with color.
    """

    def __init__(
        self,
        text: str,
        *,
        color: Color | BGColor = Color.DEFAULT,
    ) -> None:
        super().__init__(text)
        self.obj_ref.value.rich_text = Text(text).obj_ref
        self.obj_ref.value.color = color

    @property
    def color(self) -> Color | BGColor:
        return self.obj_ref.value.color

    @color.setter
    def color(self, color: Color | BGColor) -> None:
        self.obj_ref.value.color = color
        self._update_in_notion(exclude_attrs=['paragraph.children'])


class Paragraph(ColoredTextBlock[obj_blocks.Paragraph], ChildrenMixin, wraps=obj_blocks.Paragraph):
    """Paragraph block."""

    def to_markdown(self) -> str:
        return f'{self.rich_text.to_markdown()}'


class Heading(ColoredTextBlock[BT], ChildrenMixin, ABC, wraps=obj_blocks.Heading):
    """Abstract Heading block.

    Parent class of all heading block types.
    """

    def __init__(
        self,
        text: str,
        *,
        color: Color | BGColor = Color.DEFAULT,
        toggleable: bool = False,
    ) -> None:
        super().__init__(text, color=color)
        self.obj_ref.value.is_toggleable = toggleable

    @property
    def toggleable(self) -> bool:
        """Return whether the heading is toggleable."""
        return self.obj_ref.value.is_toggleable

    @toggleable.setter
    def toggleable(self, toggleable: bool) -> None:
        """Set the heading to be toggleable or not."""
        if self.toggleable and self.has_children:
            msg = 'Cannot make a toggleable heading non-toggleable if it has children.'
            raise InvalidAPIUsageError(msg)
        self.obj_ref.value.is_toggleable = toggleable
        self._update_in_notion()

    def append(self, blocks: Block | Sequence[Block], *, after: Block | None = None) -> Self:
        if not self.toggleable:
            msg = 'Cannot append blocks to a non-toggleable heading.'
            raise InvalidAPIUsageError(msg)
        return super().append(blocks, after=after)


class Heading1(Heading[obj_blocks.Heading1], wraps=obj_blocks.Heading1):
    """Heading 1 block."""

    def to_markdown(self) -> str:
        return f'## {self.rich_text.to_markdown()}'  # we use ## as # is used for page titles


class Heading2(Heading[obj_blocks.Heading2], wraps=obj_blocks.Heading2):
    """Heading 2 block."""

    def to_markdown(self) -> str:
        return f'### {self.rich_text.to_markdown()}'


class Heading3(Heading[obj_blocks.Heading3], wraps=obj_blocks.Heading3):
    """Heading 3 block."""

    def to_markdown(self) -> str:
        return f'#### {self.rich_text.to_markdown()}'


class Quote(ColoredTextBlock[obj_blocks.Quote], ChildrenMixin, wraps=obj_blocks.Quote):
    """Quote block."""

    def to_markdown(self) -> str:
        return f'> {self.rich_text.to_markdown()}\n'


class Callout(ColoredTextBlock[obj_blocks.Callout], ChildrenMixin, wraps=obj_blocks.Callout):
    """Callout block.

    !!! note

        The default icon is an electric light bulb, i.e. 💡 in case `None` is passed as icon.
        It is not possible to remove the icon, but you can change it to a different emoji or a file.
    """

    def __init__(
        self,
        text: str,
        *,
        color: Color | BGColor = Color.DEFAULT,
        icon: FileInfo | Emoji | None = None,
    ) -> None:
        super().__init__(text, color=color)
        if icon is not None:
            self.obj_ref.value.icon = icon.obj_ref

    @staticmethod
    def get_default_icon() -> Emoji:
        """Return the default icon of a callout block."""
        return Emoji('💡')

    @property
    def icon(self) -> FileInfo | Emoji:
        if (icon := self.obj_ref.value.icon) is None:
            return self.get_default_icon()
        else:
            return wrap_icon(icon)

    @icon.setter
    def icon(self, icon: FileInfo | Emoji | None) -> None:
        if icon is None:
            icon = self.get_default_icon()
        self.obj_ref.value.icon = icon.obj_ref
        self._update_in_notion()

    def to_markdown(self) -> str:
        match self.icon:
            case Emoji():
                return f'{self.icon} {self.rich_text.to_markdown()}\n'
            case FileInfo():
                return f'![icon]({self.icon.url}) {self.rich_text.to_markdown()}\n'
            case _:
                msg = f'Invalid icon type {type(self.icon)}'
                raise ValueError(msg)


class BulletedItem(ColoredTextBlock[obj_blocks.BulletedListItem], ChildrenMixin, wraps=obj_blocks.BulletedListItem):
    """Bulleted list item."""

    def to_markdown(self) -> str:
        return f'- {self.rich_text.to_markdown()}\n'


class NumberedItem(ColoredTextBlock[obj_blocks.NumberedListItem], ChildrenMixin, wraps=obj_blocks.NumberedListItem):
    """Numbered list item."""

    def to_markdown(self) -> str:
        return f'1. {self.rich_text.to_markdown()}\n'


class ToDoItem(ColoredTextBlock[obj_blocks.ToDo], ChildrenMixin, wraps=obj_blocks.ToDo):
    """ToDo list item."""

    def __init__(
        self,
        text: str,
        *,
        checked: bool = False,
        color: Color | BGColor = Color.DEFAULT,
    ) -> None:
        super().__init__(text, color=color)
        self.obj_ref.value.checked = checked

    @property
    def checked(self) -> bool:
        return self.obj_ref.to_do.checked

    @checked.setter
    def checked(self, checked: bool) -> None:
        self.obj_ref.value.checked = checked
        self._update_in_notion()

    def to_markdown(self) -> str:
        mark = 'x' if self.checked else ' '
        return f'- [{mark}] {self.rich_text.to_markdown()}\n'


class ToggleItem(ColoredTextBlock[obj_blocks.Toggle], ChildrenMixin, wraps=obj_blocks.Toggle):
    """Toggle list item."""

    def to_markdown(self) -> str:
        return f'- {self.rich_text.to_markdown()}\n'


class Divider(Block[obj_blocks.Divider], wraps=obj_blocks.Divider):
    """Divider block."""

    def to_markdown(self) -> str:  # noqa: PLR6301
        return '---\n'


class TableOfContents(Block[obj_blocks.TableOfContents], wraps=obj_blocks.TableOfContents):
    """Table of Contents block."""

    def __init__(self, *, color: Color | BGColor = Color.DEFAULT):
        super().__init__()
        self.obj_ref.value.color = color

    def to_markdown(self) -> str:  # noqa: PLR6301
        return '```{toc}\n```'


class Breadcrumb(Block[obj_blocks.Breadcrumb], wraps=obj_blocks.Breadcrumb):
    """Breadcrumb block."""

    def to_markdown(self) -> str:
        def is_page(obj: NotionEntity) -> TypeGuard[Page]:
            return obj.is_page

        return ' / '.join(ancestor.title for ancestor in self.ancestors if is_page(ancestor)) + '\n'


class Embed(Block[obj_blocks.Embed], wraps=obj_blocks.Embed):
    """Embed block."""

    def __init__(self, url: str, *, caption: str | None = None):
        super().__init__()
        self.obj_ref.value.url = url
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else None

    @property
    def url(self) -> str | None:
        """Return the URL of the embedded item."""
        return self.obj_ref.value.url

    @url.setter
    def url(self, url: str) -> None:
        self.obj_ref.value.url = url
        self._update_in_notion()

    @property
    def caption(self) -> Text:
        """Return the caption of the embedded item."""
        return Text.wrap_obj_ref(self.obj_ref.value.caption)

    @caption.setter
    def caption(self, caption: str | None) -> None:
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else []
        self._update_in_notion()

    def to_markdown(self) -> str:
        if self.url is not None:
            return f'[{self.url}]({self.url})\n'
        else:
            return ''


class Bookmark(Block[obj_blocks.Bookmark], wraps=obj_blocks.Bookmark):
    """Bookmark block."""

    def __init__(self, url: str, *, caption: str | None = None):
        super().__init__()
        self.obj_ref.value.url = url
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else None

    @property
    def url(self) -> str | None:
        """Return the URL of the bookmark."""
        return self.obj_ref.value.url

    @url.setter
    def url(self, url: str) -> None:
        self.obj_ref.value.url = url
        self._update_in_notion()

    def to_markdown(self) -> str:
        if self.url is not None:
            return f'Bookmark: [{self.url}]({self.url})\n'
        else:
            return 'Bookmark: [Add a web bookmark]()\n'  # emtpy bookmark


class LinkPreview(Block[obj_blocks.LinkPreview], wraps=obj_blocks.LinkPreview):
    """Link preview block.

    !!! warning "Not Supported"

        The `link_preview` block can only be returned as part of a response.
        The Notion API does not support creating or appending `link_preview` blocks.
    """

    def __init__(self, url: str):
        msg = 'The Notion API does not support creating or appending `link_preview` blocks.'
        raise NotImplementedError(msg)
        # ToDo: Implement this when the API supports it
        # super().__init__()
        # self.obj_ref.value.url = url

    @property
    def url(self) -> str | None:
        return self.obj_ref.value.url

    def to_markdown(self) -> str:
        if self.block_url is not None:
            return f'Link preview: [{self.url}]({self.url})\n'
        return super().to_markdown()


class Equation(Block[obj_blocks.Equation], wraps=obj_blocks.Equation):
    """Equation block.

    LaTeX equation in display mode, e.g. `$$ \\mathrm{E=mc^2} $$`, but without the `$$` signs.
    """

    def __init__(self, latex: str):
        super().__init__()
        self.obj_ref.value.expression = latex

    @property
    def latex(self) -> str:
        if (expr := self.obj_ref.equation.expression) is not None:
            return expr.rstrip()
        else:
            return ''

    @latex.setter
    def latex(self, latex: str) -> None:
        self.obj_ref.value.expression = latex
        self._update_in_notion()

    def to_markdown(self) -> str:
        return f'$$\n{self.latex}\n$$\n'


FT = TypeVar('FT', bound=obj_blocks.FileBase)


class FileBaseBlock(Block[FT], ABC, wraps=obj_blocks.FileBase):
    """Abstract Block for file-based blocks.

    Parent class of all file-based block types.
    """

    def __init__(
        self,
        url: str,
        *,
        caption: str | None = None,
        name: str | None = None,
    ):
        super().__init__()
        file_info = FileInfo(url=url, name=name, caption=caption)
        self.obj_ref.value = file_info.obj_ref

    @property
    def file_info(self) -> FileInfo:
        """Return the file information of this block as a copy."""
        if isinstance(file_obj := self.obj_ref.value, objs.FileObject):
            return FileInfo.wrap_obj_ref(file_obj.model_copy(deep=True))
        else:
            msg = f'Unknown file type {type(file_obj)}'
            raise ValueError(msg)

    @property
    def url(self) -> str:
        return self.file_info.url

    @property
    def caption(self) -> Text:
        return self.file_info.caption

    @caption.setter
    def caption(self, caption: str | None) -> None:
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else []
        self._update_in_notion()


class File(FileBaseBlock[obj_blocks.File], wraps=obj_blocks.File):
    """File block.

    !!! note

        Only the caption and name can be modified, the URL is read-only.
        Note that only the file name not the file suffix can be modified,
        the suffix is determined initially by the url.
    """

    def __init__(
        self,
        name: str,
        url: str,
        *,
        caption: str | None = None,
    ):
        super().__init__(url, caption=caption, name=name)

    @property
    def name(self) -> str:
        name = self.file_info.name
        return name if name is not None else ''

    @name.setter
    def name(self, name: str) -> None:
        self.obj_ref.value.name = name
        self._update_in_notion()

    def to_markdown(self) -> str:
        md = f'[📎 {self.name}]({self.url})\n'
        if self.caption:
            md += f'{self.caption.to_markdown()}\n'
        return md


class Image(FileBaseBlock[obj_blocks.Image], wraps=obj_blocks.Image):
    """Image block.

    !!! note

        Only the caption can be modified, the URL is read-only.
    """

    def __init__(self, url: str, *, caption: str | None = None):
        super().__init__(url, caption=caption)

    def to_markdown(self) -> str:
        alt = self.url.rsplit('/').pop()
        caption = self.caption.to_plain_text()
        if caption:
            return f'<figure><img src="{self.url}" alt="{alt}" /><figcaption>{caption}</figcaption></figure>\n'
        else:
            return f'![{alt}]({self.url})\n'


class Video(FileBaseBlock[obj_blocks.Video], wraps=obj_blocks.Video):
    """Video block.

    !!! note

        Only the caption can be modified, the URL is read-only.
    """

    def to_markdown(self) -> str:
        mime_type, _ = mimetypes.guess_type(self.url)
        vtype = f' type="{mime_type}"' if mime_type else ''
        md = f'<video width="320" height="240" controls><source src="{self.url}"{vtype}></video>\n'
        return md


class PDF(FileBaseBlock[obj_blocks.PDF], wraps=obj_blocks.PDF):
    """PDF block.

    !!! note

        Only the caption can be modified, the URL is read-only.
    """

    def to_markdown(self) -> str:
        name = self.url.rsplit('/').pop()
        md = f'[📖 {name}]({self.url})\n'
        if self.caption:
            md += f'{self.caption.to_markdown()}\n'
        return md


class ChildPage(Block[obj_blocks.ChildPage], wraps=obj_blocks.ChildPage):
    """Child page block.

    !!! note

        To create a child page block, create a new page with the corresponding parent.
        This block is used only internally.
    """

    def __init__(self):
        msg = 'To create a child page block, create a new page with the corresponding parent.'
        raise InvalidAPIUsageError(msg)

    @property
    def title(self) -> str:
        """Return the title of the child page."""
        return self.obj_ref.child_page.title

    @property
    def page(self) -> Page:
        """Return the actual Page object."""
        sess = get_active_session()
        return sess.get_page(self.obj_ref.id)

    def to_markdown(self) -> str:  # noqa: PLR6301
        msg = '`ChildPage` is only used internally, work with a proper `Page` instead.'
        raise InvalidAPIUsageError(msg)


class ChildDatabase(Block[obj_blocks.ChildDatabase], wraps=obj_blocks.ChildDatabase):
    """Child database block.

    !!! note

        To create a child database block as an end-user, create a new database with the corresponding parent.
        This block is used only internally.
    """

    def __init__(self):
        msg = 'To create a child database block, create a new database with the corresponding parent.'
        raise InvalidAPIUsageError(msg)

    @property
    def title(self) -> str:
        """Return the title of the child database"""
        return self.obj_ref.child_database.title

    @property
    def db(self) -> Database:
        """Return the actual Database object."""
        sess = get_active_session()
        return sess.get_db(self.obj_ref.id)

    def to_markdown(self) -> str:  # noqa: PLR6301
        msg = '`ChildDatabase` is only used internally, work with a proper `Database` instead.'
        raise InvalidAPIUsageError(msg)


class Column(Block[obj_blocks.Column], ChildrenMixin, wraps=obj_blocks.Column):
    """Column block."""

    def __init__(self):
        msg = 'Column blocks cannot be created directly. Use `Columns` instead.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        mds = []
        for block in self.children:
            mds.append(block.to_markdown())
        return '\n'.join(mds)


class Columns(Block[obj_blocks.ColumnList], ChildrenMixin, wraps=obj_blocks.ColumnList):
    """Columns block."""

    def __init__(self, n_columns: int):
        """Create a new `Columns` block with the given number of columns."""
        super().__init__()
        self.obj_ref.column_list.children = [obj_blocks.Column.build() for _ in range(n_columns)]

    def __getitem__(self, index: int) -> Column:
        return cast(Column, self.children[index])

    def add_column(self, index: int = -1) -> Self:
        """Add a new column to this block of columns."""
        if 1 <= index <= len(self.children):
            new_col = Column.wrap_obj_ref(obj_blocks.Column.build())
            super().append(new_col, after=self.children[index - 1])
            return self
        else:
            msg = f'Columns can only be inserted from index 1 to {len(self.children)}.'
            raise IndexError(msg)

    def append(self, blocks: Block | Sequence[Block], *, after: Block | None = None) -> Self:  # noqa: PLR6301
        msg = 'Use `add_column` to append a new column.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        cols = []
        for i, block in enumerate(self.children):
            md = md_comment(f'column {i + 1}')
            cols.append(md + block.to_markdown())
        return '\n'.join(cols)


class TableRow(tuple[Text, ...], Block[obj_blocks.TableRow], wraps=obj_blocks.TableRow):
    """Table row block behaving like a tuple."""

    def __new__(cls, *cells: str) -> TableRow:
        return tuple.__new__(cls, cells)

    def __init__(self, *cells: str):
        super().__init__(n_cells=len(cells))
        for idx, cell in enumerate(cells):
            self.obj_ref.table_row.cells[idx] = Text(cell).obj_ref

    @classmethod
    def wrap_obj_ref(cls, obj_ref: obj_blocks.TableRow, /) -> TableRow:
        row = TableRow(*[Text.wrap_obj_ref(cell) for cell in obj_ref.table_row.cells])
        row.obj_ref = obj_ref
        return row

    def to_markdown(self) -> str:
        return ' | '.join(self)


class Table(Block[obj_blocks.Table], ChildrenMixin, wraps=obj_blocks.Table):
    """Table block."""

    def __init__(self, n_rows: int, n_cols: int, *, header_col: bool = False, header_row: bool = False):
        super().__init__()
        self.obj_ref.table.table_width = n_cols
        self.obj_ref.table.has_column_header = header_row
        self.obj_ref.table.has_row_header = header_col
        self.obj_ref.table.children = [obj_blocks.TableRow.build(n_cols) for _ in range(n_rows)]
        self.obj_ref.has_children = bool(self.obj_ref.table.children)

    def _check_index(self, index: int | tuple[int, int]) -> tuple[int, int | None]:
        if isinstance(index, tuple):
            row_idx, col_idx = index
        else:
            row_idx, col_idx = index, None

        if not (0 <= row_idx < len(self.children)):
            msg = 'Row index out of range'
            raise IndexError(msg)

        if col_idx is not None and not (0 <= col_idx < self.width):
            msg = 'Column index out of range'
            raise IndexError(msg)

        return row_idx, col_idx

    def __str__(self) -> str:
        return self.to_markdown()

    @overload
    def __getitem__(self, index: int) -> TableRow: ...

    @overload
    def __getitem__(self, index: tuple[int, int]) -> Text: ...

    def __getitem__(self, index: int | tuple[int, int]) -> Text | TableRow:
        row_idx, col_idx = self._check_index(index)
        row = self.children[row_idx]
        return row if col_idx is None else row[col_idx]

    def __setitem__(
        self,
        index: int | tuple[int, int],
        value: str | Sequence[str],
    ) -> None:
        row_idx, col_idx = self._check_index(index)
        row = self.children[row_idx]
        row_obj = row.obj_ref
        if col_idx is None:
            if not isinstance(value, Sequence) or len(value) != self.width:
                msg = 'Value is no sequence or its length does not match the width of the table.'
                raise ValueError(msg)
            for idx, item in enumerate(value):
                row_obj.table_row.cells[idx] = Text(item).obj_ref
        elif isinstance(value, str):
            row_obj.table_row.cells[col_idx] = Text(value).obj_ref
        else:
            msg = 'A single value is needed to set a single cell.'
            raise ValueError(msg)
        row._update_in_notion()
        if self._children is not None:  # update cache
            self._children[row_idx] = TableRow.wrap_obj_ref(row_obj)  # needed to call __new__ on the TableRow

    @property
    def children(self) -> tuple[TableRow, ...]:
        """Return all rows of the table."""
        return tuple(cast(TableRow, block) for block in super().children)

    @property
    def width(self) -> int:
        """Return the width, i.e. number of columns, of the table."""
        return self.obj_ref.table.table_width

    @property
    def shape(self) -> tuple[int, int]:
        """Return the shape of the table."""
        n_cols = self.width
        n_rows = len(self.children)
        return n_rows, n_cols

    @property
    def has_header_row(self) -> bool:
        """Return whether the table has a header row."""
        return self.obj_ref.table.has_column_header

    @has_header_row.setter
    def has_header_row(self, header_row: bool) -> None:
        self.obj_ref.table.has_column_header = header_row
        self._update_in_notion(exclude_attrs=['table.table_width', 'table.children'])

    @property
    def has_header_col(self) -> bool:
        """Return whether the table has a header column."""
        return self.obj_ref.table.has_row_header

    @has_header_col.setter
    def has_header_col(self, header_col: bool) -> None:
        self.obj_ref.table.has_row_header = header_col
        self._update_in_notion(exclude_attrs=['table.table_width', 'table.children'])

    def to_markdown(self) -> str:
        """Return the table as Markdown."""
        headers = 'firstrow' if self.has_header_row else [''] * self.width
        return tabulate(self.children, headers, tablefmt='github') + '\n'

    def insert_row(self, index: int, values: Sequence[str]) -> Self:
        """Insert a new row at the given index."""
        if not (1 <= index <= len(self.children)):
            msg = f'Columns can only be inserted from index 1 to {len(self.children)}.'
            raise IndexError(msg)
        if len(values) != self.width:
            msg = 'Length of passed value does not match width of table.'
            raise ValueError(msg)

        self.append(TableRow(*values), after=self.children[index - 1])
        return self

    def append_row(self, values: Sequence[str]) -> Self:
        """Append a new row to the table."""
        return self.insert_row(len(self.children), values)


class LinkToPage(Block[obj_blocks.LinkToPage], wraps=obj_blocks.LinkToPage):
    """Link to page block.

    !!! note
        Updating a link to page block is not supported by the Notion API.
        Use `.replace(new_block)` instead.
    """

    def __init__(self, page: Page):
        super().__init__()
        self.obj_ref.link_to_page = objs.PageRef.build(page.obj_ref)

    @property
    def page(self) -> Page:
        session = get_active_session()
        return session.get_page(objs.get_uuid(self.obj_ref.link_to_page))

    @page.setter
    def page(self, page: Page) -> None:  # noqa: PLR6301
        # Updating a link to page block is not supported by the API and returns the current one.
        # self.obj_ref.link_to_page = objs.PageRef.build(page.obj_ref)
        # self._update_in_notion()
        msg = 'Updating a link to page block is not supported by the API. Use `.replace(new_block)` instead.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        return f'[**↗️ <u>{self.page.title}</u>**]({self.page.url})\n'


class SyncedBlock(Block[obj_blocks.SyncedBlock], ChildrenMixin, wraps=obj_blocks.SyncedBlock):
    """Synced block - either original or synced."""

    def __init__(self, blocks: Block | Sequence[Block]):
        """Create the original synced block."""
        super().__init__()
        blocks = [blocks] if isinstance(blocks, Block) else blocks
        self.obj_ref.synced_block.children = [block.obj_ref for block in blocks]

    @property
    def is_original(self) -> bool:
        """Return if this block is the original block."""
        return self.obj_ref.synced_block.synced_from is None

    @property
    def is_synced(self) -> bool:
        """Return if this block is synced from another block."""
        return not self.is_original

    def get_original(self) -> SyncedBlock:
        """Return the original block."""
        if self.is_original:
            return self
        elif (synced_from := self.obj_ref.synced_block.synced_from) is not None:
            session = get_active_session()
            return cast(SyncedBlock, session.get_block(objs.get_uuid(synced_from)))
        else:
            msg = 'Unknown synced block, neither original nor synced!'
            raise RuntimeError(msg)

    def create_synced(self) -> SyncedBlock:
        """Return the synced block for appending."""
        if not self.in_notion:
            msg = 'Cannot create a synced block for a block that is not in Notion. Append first!'
            raise RuntimeError(msg)

        if not self.is_original:
            msg = 'Cannot create a synced block for a block that is already synced.'
            raise RuntimeError(msg)

        obj = obj_blocks.SyncedBlock.build()
        obj.synced_block.synced_from = obj_blocks.BlockRef.build(self.obj_ref)
        return self.wrap_obj_ref(obj)

    def to_markdown(self, *, with_comment: bool = True) -> str:
        if self.is_original:
            md = md_comment('original block') if with_comment else ''
            md += '\n'.join([child.to_markdown() for child in self.children])
        else:
            md = md_comment('synced block') if with_comment else ''
            md += self.get_original().to_markdown(with_comment=False)
        return md


class Template(TextBlock[obj_blocks.Template], ChildrenMixin, wraps=obj_blocks.Template):
    """Template block.

    !!! warning "Deprecated"

        As of March 27, 2023 creation of template blocks will no longer be supported.
    """

    def __init__(self):
        msg = 'A template block cannot be created by a user.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        return f'<button type="button">{self.rich_text.to_markdown()}</button>\n'


class Unsupported(Block[obj_blocks.UnsupportedBlock], wraps=obj_blocks.UnsupportedBlock):
    """Unsupported block in the API."""

    def __init__(self):
        msg = 'An unsupported block cannot be created by a user.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:  # noqa: PLR6301
        return '<kbd>Unsupported block</kbd>\n'



"""Functionality for working with Notion databases."""

from __future__ import annotations

from typing import cast

from pydantic import BaseModel
from typing_extensions import Self

from ultimate_notion.blocks import ChildrenMixin, DataObject
from ultimate_notion.core import get_active_session, get_repr
from ultimate_notion.file import Emoji, FileInfo
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.page import Page
from ultimate_notion.query import Query
from ultimate_notion.rich_text import Text, camel_case, snake_case
from ultimate_notion.schema import Property, PropertyType, PropertyValue, ReadOnlyPropertyError, Schema, SchemaError
from ultimate_notion.view import View


class Database(DataObject[obj_blocks.Database], wraps=obj_blocks.Database):
    """A Notion database.

    This object always represents an original database, not a linked database.

    API reference: https://developers.notion.com/docs/working-with-databases
    """

    _schema: type[Schema] | None = None

    def __str__(self):
        if self.title:
            return str(self.title)
        else:
            return 'Untitled'

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically"""
        return str(self)

    def __repr__(self) -> str:
        return get_repr(self)

    @property
    def url(self) -> str:
        """Return the URL of this database."""
        return self.obj_ref.url

    @property
    def title(self) -> str | Text:
        """Return the title of this database as rich text."""
        title = self.obj_ref.title
        # `str` added as return value but always RichText returned, which inherits from str.
        return Text.wrap_obj_ref(title)

    @title.setter
    def title(self, text: str | Text):
        """Set the title of this database"""
        if not isinstance(text, Text):
            text = Text.from_plain_text(text)
        session = get_active_session()
        session.api.databases.update(self.obj_ref, title=text.obj_ref)

    @property
    def description(self) -> Text:
        """Return the description of this database as rich text."""
        desc = self.obj_ref.description
        return Text.wrap_obj_ref(desc)

    @description.setter
    def description(self, text: str | Text):
        """Set the description of this database."""
        if not isinstance(text, Text):
            text = Text.from_plain_text(text)
        session = get_active_session()
        session.api.databases.update(self.obj_ref, description=text.obj_ref)

    @property
    def icon(self) -> FileInfo | Emoji | None:
        """Return the icon of this database as file or emoji."""
        match self.obj_ref.icon:
            case objs.FileObject() as icon:
                return FileInfo.wrap_obj_ref(icon)
            case objs.EmojiObject() as icon:
                return Emoji.wrap_obj_ref(icon)
            case None:
                return None
            case _:
                msg = f'Unknown icon object of {type(self.obj_ref.icon)}'
                raise RuntimeError(msg)

    @property
    def cover(self) -> FileInfo | None:
        """Return the cover of this database as file."""
        cover = self.obj_ref.cover
        return FileInfo.wrap_obj_ref(cover) if cover is not None else None

    @property
    def is_wiki(self) -> bool:
        """Return whether the database is a wiki."""
        # ToDo: Implement using the verification property
        raise NotImplementedError

    @property
    def is_db(self) -> bool:
        """Return whether the object is a database."""
        return True

    def _reflect_schema(self, obj_ref: obj_blocks.Database) -> type[Schema]:
        """Reflection about the database schema."""
        title = str(self)
        cls_name = f'{camel_case(title)}Schema'
        attrs = {
            snake_case(k): Property(k, cast(PropertyType, PropertyType.wrap_obj_ref(v)))
            for k, v in obj_ref.properties.items()
        }
        schema: type[Schema] = type(cls_name, (Schema,), attrs, db_title=title)
        schema.bind_db(self)
        return schema

    def _set_schema(self, schema: type[Schema], *, during_init: bool):
        """Set a custom schema in order to change the Python variables names."""
        self.schema.assert_consistency_with(schema, during_init=during_init)
        schema.bind_db(self)
        self._schema = schema

    @property
    def schema(self) -> type[Schema]:
        """Schema of the database."""
        if not self._schema:
            self._schema = self._reflect_schema(self.obj_ref)
        return self._schema

    @schema.setter
    def schema(self, schema: type[Schema]):
        """Set a custom schema in order to change the Python variables names."""
        self._set_schema(schema, during_init=False)

    @property
    def is_inline(self) -> bool:
        """Return whether the database is inline."""
        return self.obj_ref.is_inline

    def delete(self) -> Self:
        """Delete this database."""
        if not self.is_deleted:
            session = get_active_session()
            session.api.databases.delete(self.obj_ref)
            self._delete_me_from_parent()
        return self

    def restore(self) -> Self:
        """Restore this database."""
        if self.is_deleted:
            session = get_active_session()
            session.api.databases.restore(self.obj_ref)
            if isinstance(self.parent, ChildrenMixin):
                self.parent._children = None  # this forces a new retrieval of children next time
        return self

    def reload(self) -> Self:
        """Reload this database."""
        session = get_active_session()
        self.obj_ref = session.api.databases.retrieve(self.obj_ref.id)
        self.schema._set_obj_refs()
        return self

    @property
    def query(self) -> Query:
        """Return a Query object to build and execute a database query."""
        return Query(database=self)

    def get_all_pages(self) -> View:
        """Retrieve all pages and return a view."""
        return self.query.execute()

    def __len__(self) -> int:
        """Return the number of pages in this database."""
        return len(self.get_all_pages())

    @property
    def is_empty(self) -> bool:
        """Return whether the database is empty."""
        return len(self) == 0

    def __bool__(self) -> bool:
        """Overwrite default behaviour."""
        msg = 'Use .is_empty instead of bool(db) to check if a database is empty.'
        raise RuntimeError(msg)

    def pydantic_model(self) -> BaseModel:
        """Return a Pydantic model for this database."""
        # Check https://github.com/ultimate-notion/ultimate-notion/issues/32 for details
        raise NotImplementedError

    def create_page(self, **kwargs) -> Page:
        """Create a page with properties according to the schema within the corresponding database."""

        # ToDo: let pydantic_model check the kwargs and raise an error if something is wrong
        #       this would also generate a nice error message for the user!
        schema_kwargs = {prop.attr_name: prop for prop in self.schema.get_props()}
        if not set(kwargs).issubset(set(schema_kwargs)):
            add_kwargs = set(kwargs) - set(schema_kwargs)
            msg = f"kwargs {', '.join(add_kwargs)} not defined in schema"
            raise SchemaError(msg)

        schema_dct = {}
        for kwarg, value in kwargs.items():
            prop = schema_kwargs[kwarg]
            prop_value_cls = prop.type.prop_value  # map schema to page property
            # ToDo: Check at that point in case of select option if the option is already defined in Schema!

            if prop_value_cls.readonly:
                raise ReadOnlyPropertyError(prop)

            prop_value = value if isinstance(value, PropertyValue) else prop_value_cls(value)

            schema_dct[schema_kwargs[kwarg].name] = prop_value.obj_ref

        session = get_active_session()
        page_obj = session.api.pages.create(parent=self.obj_ref, properties=schema_dct)
        page = Page.wrap_obj_ref(page_obj)
        session.cache[page.id] = page
        return page

    def to_markdown(self) -> str:
        """Return the content of this database as Markdown."""
        raise NotImplementedError

    def _to_markdown(self) -> str:
        """Return the reference to this database as Markdown."""
        return f'[**🗄️ {self.title}**]({self.block_url})\n'



"""Ultimate Notion provides a pythonic, high-level API for Notion.

Notion-API: https://developers.notion.com/reference/intro
"""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version('ultimate-notion')
except PackageNotFoundError:  # pragma: no cover
    __version__ = 'unknown'
finally:
    del version, PackageNotFoundError

# ToDo: Activate when beartype 0.19.1 fixes https://github.com/beartype/beartype/issues/444
# from beartype.claw import beartype_this_package

from ultimate_notion.blocks import (
    PDF,
    AnyBlock,
    Bookmark,
    Breadcrumb,
    BulletedItem,
    Callout,
    Code,
    Column,
    Columns,
    Divider,
    Embed,
    Equation,
    File,
    Heading1,
    Heading2,
    Heading3,
    Image,
    LinkPreview,
    LinkToPage,
    NumberedItem,
    Paragraph,
    Quote,
    SyncedBlock,
    Table,
    TableOfContents,
    TableRow,
    ToDoItem,
    ToggleItem,
    Video,
)
from ultimate_notion.core import get_active_session
from ultimate_notion.database import Database, ReadOnlyPropertyError
from ultimate_notion.file import Emoji, FileInfo
from ultimate_notion.obj_api.enums import AggFunc, BGColor, CodeLang, Color, NumberFormat, VState
from ultimate_notion.option import Option, OptionNS
from ultimate_notion.page import Page
from ultimate_notion.query import Condition, prop
from ultimate_notion.rich_text import join, math, mention, text
from ultimate_notion.schema import Property, PropType, Schema, SchemaError, SelfRef
from ultimate_notion.session import Session
from ultimate_notion.user import User
from ultimate_notion.utils import SList

__all__ = [
    'PDF',
    'AggFunc',
    'AnyBlock',  # for type hinting
    'BGColor',
    'Bookmark',
    'Breadcrumb',
    'BulletedItem',
    'Callout',
    'Code',
    'CodeLang',
    'Color',
    'Column',
    'Columns',
    'Condition',  # for type hinting
    'Database',
    'Divider',
    'Embed',
    'Emoji',
    'Equation',
    'File',
    'FileInfo',
    'Heading1',
    'Heading2',
    'Heading3',
    'Image',
    'LinkPreview',
    'LinkToPage',
    'NumberFormat',
    'NumberedItem',
    'Option',
    'OptionNS',
    'Page',
    'Paragraph',
    'PropType',
    'Property',
    'Quote',
    'ReadOnlyPropertyError',
    'SList',
    'Schema',
    'SchemaError',
    'SelfRef',
    'Session',
    'SyncedBlock',
    'Table',
    'TableOfContents',
    'TableRow',
    'ToDoItem',
    'ToggleItem',
    'User',
    'VState',
    'Video',
    '__version__',
    'get_active_session',
    'join',
    'math',
    'mention',
    'prop',
    'text',
]


# ToDo: Activate when beartype 0.19.1 fixes https://github.com/beartype/beartype/issues/444
# activate fast near-real-time hybrid runtime-static type-checking
# beartype_this_package()



"""Functionality for general Notion objects like texts, files, options, etc."""

from __future__ import annotations

from ultimate_notion.core import Wrapper, get_repr
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.enums import Color
from ultimate_notion.rich_text import Text


class Option(Wrapper[objs.SelectOption], wraps=objs.SelectOption):
    """Option for select & multi-select property."""

    def __init__(self, name: str, *, color: Color | str = Color.DEFAULT):
        if isinstance(color, str):
            color = Color(color)
        super().__init__(name, color=color)

    @property
    def id(self) -> str:
        """ID of the option."""
        return self.obj_ref.id

    @property
    def name(self) -> str:
        """Name of the option."""
        return self.obj_ref.name

    def description(self) -> str:
        """Description of the option."""
        if desc := self.obj_ref.description:
            return Text.wrap_obj_ref(desc)
        else:
            return ''

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Option):
            # We compare only the name as the id is not set for new options
            res = self.name == other.name
        elif isinstance(other, str):
            res = self.name == other
        elif other is None:
            res = False
        else:
            msg = f'Cannot compare Option with types {type(other)}'
            raise RuntimeError(msg)
        return res

    def __hash__(self) -> int:
        return super().__hash__()


class OptionNSType(type):
    """Metaclass to implement `len` for type `OptionNS` itself, not an instance of it."""

    def __len__(cls):
        return len(cls.to_list())


class OptionNS(metaclass=OptionNSType):
    """Option namespace to simplify working with (Multi-)Select options."""

    @classmethod
    def to_list(cls) -> list[Option]:
        """Convert the enum to a list as needed by the (Multi)Select property types."""
        return [
            getattr(cls, var) for var in cls.__dict__ if not var.startswith('__') and not callable(getattr(cls, var))
        ]


class OptionGroup(Wrapper[objs.SelectGroup], wraps=objs.SelectGroup):
    """Group of options for status property."""

    _options: dict[str, Option]  # holds all possible options

    @classmethod
    def wrap_obj_ref(cls, obj_ref, /, *, options: list[Option] | None = None) -> OptionGroup:
        """Convienence constructor for the group of options."""
        obj = super().wrap_obj_ref(obj_ref)
        options = [] if options is None else options
        obj._options = {option.id: option for option in options}
        return obj

    @property
    def name(self) -> str:
        """Name of the option group."""
        return self.obj_ref.name

    @property
    def options(self) -> list[Option]:
        """Options within this option group."""
        return [self._options[opt_id] for opt_id in self.obj_ref.option_ids]

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.name



"""Utilities to synchronize other services with Notion."""

from __future__ import annotations

import asyncio
import logging
import pickle  # noqa: S403
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, TypeAlias

from pydantic import BaseModel, Field
from typing_extensions import Self

from ultimate_notion.config import get_cfg

_logger = logging.getLogger(__name__)

ID: TypeAlias = str


class ConflictMode(str, Enum):
    """Conflict resolution modes."""

    NEWER = 'newer'
    NOTION = 'notion'
    OTHER = 'other'  # any other service than Notion
    ERROR = 'error'


class State(BaseModel):
    """The state of a sync task.

    The state holds the synced objects and their attributes as a dictionary of Notion attributes."""

    ids: dict[ID, ID] = Field(default_factory=dict)
    """Maps Notion object ids to other object ids."""
    objs: dict[ID, dict[str, Any]] = Field(default_factory=dict)
    """Dictionary of Notion objects synced with other service and indexed by their ids."""


class SyncTask(ABC):
    """A task to be performed during a sync.

    This is an abstract base class to allow for different kinds of sync tasks.
    """

    _run_every_secs: float | None = None
    _in_total_times: int | None = None

    def __init__(
        self, name: str, attr_map: dict[str, str], conflict_mode: ConflictMode | str = ConflictMode.NEWER
    ) -> None:
        if isinstance(conflict_mode, str):
            conflict_mode = ConflictMode(conflict_mode)

        self.name = name
        self.attr_map = attr_map
        self.conflict_mode = conflict_mode
        self.state_path = get_cfg().ultimate_notion.sync_state_dir / f'{name}.pickle'
        super().__init__()

    def schedule(self) -> Self:
        """Apply the task."""
        all_tasks.add(self)
        return self

    def run_every(self, hours: int = 0, minutes: int = 0, seconds: int = 0) -> Self:
        """Schedule the task to run every so many seconds."""
        self._run_every_secs = timedelta(hours=hours, minutes=minutes, seconds=seconds).total_seconds()
        return self

    def run_once(self) -> Self:
        self._run_every_secs = None
        return self

    def in_total(self, times: int) -> Self:
        """Schedule the task to run a total of so many times."""
        if times <= 0:
            msg = 'times must be positive'
            raise ValueError(msg)
        self._in_total_times = times
        return self

    @abstractmethod
    def get_notion_objects(self) -> list[Any]:
        """Get all Notion objects to sync."""
        raise NotImplementedError()

    @abstractmethod
    def get_other_objects(self) -> list[Any]:
        """Get all other objects to sync."""
        raise NotImplementedError()

    @abstractmethod
    def notion_timestamp(self, obj: Any) -> datetime:
        """Get the timestamp of the Notion object."""
        raise NotImplementedError()

    @abstractmethod
    def other_timestamp(self, obj: Any) -> datetime:
        """Get the timestamp of the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_id(self, obj: Any) -> ID:
        """Get the id of the Notion object."""
        raise NotImplementedError()

    @abstractmethod
    def other_id(self, obj: Any) -> ID:
        """Get the id of the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_hash(self, obj: Any) -> str:
        """Get the hash of the Notion object for object mapping/linking."""
        raise NotImplementedError()

    @abstractmethod
    def other_hash(self, obj: Any) -> str:
        """Get the hash of the other object for object mapping/linking."""
        raise NotImplementedError()

    @abstractmethod
    def notion_to_dict(self, obj: Any) -> dict[str, Any]:
        """Convert a Notion object to a dictionary."""
        raise NotImplementedError()

    @abstractmethod
    def other_to_dict(self, obj: Any) -> dict[str, Any]:
        """Convert another object to a dictionary."""
        raise NotImplementedError()

    @abstractmethod
    def notion_update_obj(self, obj: Any, attr: str, value: Any) -> None:
        """Set an attribute of the Notion object, e.g. page."""
        raise NotImplementedError()

    @abstractmethod
    def other_update_obj(self, obj: Any, attr: str, value: Any) -> None:
        """Set an attribute of the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_delete_obj(self, obj: Any) -> None:
        """Delete the page."""
        raise NotImplementedError()

    @abstractmethod
    def other_delete_obj(self, obj: Any) -> None:
        """Delete the other object."""
        raise NotImplementedError()

    @abstractmethod
    def notion_create_obj(self, **kwargs: Any) -> Any:
        """Create a new page."""
        raise NotImplementedError()

    @abstractmethod
    def other_create_obj(self, **kwargs: Any) -> Any:
        """Create a new other object."""
        raise NotImplementedError()

    def sync_notion_deleted(self, state: State, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Sync an object in the state that was deleted in Notion."""
        for notion_id, other_id in state.ids.copy().items():
            if notion_id not in notion_objs:
                if other_id in other_objs:
                    _logger.debug(f'Deleting other object with id {other_id}')
                    self.other_delete_obj(other_objs[other_id])
                    del other_objs[other_id]
                del state.ids[notion_id]
                del state.objs[notion_id]
        return state

    def sync_other_deleted(self, state: State, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Sync an object in the state that was deleted in the other service."""
        for notion_id, other_id in state.ids.copy().items():
            if other_id not in other_objs:
                if notion_id in notion_objs:
                    _logger.debug(f'Deleting Notion object with id {notion_id}')
                    self.notion_delete_obj(notion_objs[notion_id])
                    del notion_objs[notion_id]
                del state.ids[notion_id]
                del state.objs[notion_id]
        return state

    def sync_notion_created(self, state: State, notion_objs: dict[ID, Any]) -> State:
        """Sync an object not in the state and created in Notion."""
        for notion_id, notion_obj in notion_objs.items():
            if notion_id not in state.objs:
                notion_obj_dct = self.notion_to_dict(notion_obj)
                other_obj_dct = {
                    other_attr: notion_obj_dct[notion_attr] for notion_attr, other_attr in self.attr_map.items()
                }
                _logger.debug(f'Creating other object with attributes: {other_obj_dct}')
                other_obj = self.other_create_obj(**other_obj_dct)

                if self.other_to_dict(other_obj) != other_obj_dct:
                    msg = 'The other object created by Notion does not match the expected object.'
                    raise RuntimeError(msg)

                other_id = self.other_id(other_obj)
                state.ids[notion_id] = other_id
                state.objs[notion_id] = notion_obj_dct
        return state

    def sync_other_created(self, state: State, other_objs: dict[ID, Any]) -> State:
        """Sync an object not in the state and created in other service."""
        for other_id, other_obj in other_objs.items():
            if other_id not in state.ids.values():
                other_obj_dct = self.other_to_dict(other_obj)
                notion_obj_dct = {
                    notion_attr: other_obj_dct[other_attr] for notion_attr, other_attr in self.attr_map.items()
                }
                _logger.debug(f'Creating Notion object with attributes: {notion_obj_dct}')
                notion_obj = self.notion_create_obj(**notion_obj_dct)

                if self.notion_to_dict(notion_obj) != notion_obj_dct:
                    msg = 'The Notion object created by other service does not match the expected object.'
                    raise RuntimeError(msg)

                notion_id = self.notion_id(notion_obj)
                state.ids[notion_id] = other_id
                state.objs[notion_id] = notion_obj_dct
        return state

    def resolve_conflict(self, notion_obj: Any, other_obj: Any, notion_attr: str, other_attr: str) -> Any:
        """Resolve a conflict between two objects on an attribute."""
        _logger.debug(
            f'Resolving conflict on attribute {notion_attr} of Notion object {self.notion_id(notion_obj)}'
            f'and attribute {other_attr} of other object {self.other_id(other_obj)}.'
        )
        notion_obj_dct, other_obj_dct = self.notion_to_dict(notion_obj), self.other_to_dict(other_obj)

        if self.conflict_mode == ConflictMode.NOTION:
            _logger.debug(f'Updating other object with attribute {other_attr} to {notion_obj_dct[notion_attr]}')
            self.other_update_obj(other_obj, other_attr, notion_obj_dct[notion_attr])
            return notion_obj_dct[notion_attr]
        elif self.conflict_mode == ConflictMode.OTHER:
            _logger.debug(f'Updating Notion object with attribute {notion_attr} to {other_obj_dct[other_attr]}')
            self.notion_update_obj(notion_obj, notion_attr, other_obj_dct[other_attr])
            return other_obj_dct[other_attr]
        elif self.conflict_mode == ConflictMode.NEWER:
            if self.notion_timestamp(notion_obj) > self.other_timestamp(other_obj):
                _logger.debug(f'Updating other object with attribute {other_attr} to {notion_obj_dct[notion_attr]}')
                self.other_update_obj(other_obj, other_attr, notion_obj_dct[notion_attr])
                return notion_obj_dct[notion_attr]
            else:
                _logger.debug(f'Updating Notion object with attribute {notion_attr} to {other_obj_dct[other_attr]}')
                self.notion_update_obj(notion_obj, notion_attr, other_obj_dct[other_attr])
                return other_obj_dct[other_attr]
        else:
            msg = f'Unknown conflict mode {self.conflict_mode}'
            raise RuntimeError(msg)

    def initial_sync(self, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Make the initial state.

        This is a two-way sync, i.e. the objects are compared and the differences are resolved.
        """
        _logger.debug('Performing initial sync...')
        state = State()
        notion_hashes = {self.notion_hash(obj): obj_id for obj_id, obj in notion_objs.items()}
        other_hashes = {self.other_hash(obj): obj_id for obj_id, obj in other_objs.items()}

        for obj_hash in notion_hashes.keys() & other_hashes.keys():
            notion_id, other_id = notion_hashes[obj_hash], other_hashes[obj_hash]

            state.ids[notion_id] = other_id
            state_obj = state.objs[notion_id] = {}

            notion_obj, other_obj = notion_objs[notion_id], other_objs[other_id]
            notion_obj_dct, other_obj_dct = self.notion_to_dict(notion_obj), self.other_to_dict(other_obj)

            for notion_attr, other_attr in self.attr_map.items():
                if notion_obj_dct[notion_attr] == other_obj_dct[other_attr]:
                    state_obj[notion_attr] = notion_obj_dct[notion_attr]
                else:
                    state_obj[notion_attr] = self.resolve_conflict(notion_obj, other_obj, notion_attr, other_attr)

        return state

    def sync_state_changes(self, state: State, notion_objs: dict[ID, Any], other_objs: dict[ID, Any]) -> State:
        """Sync changes with respect to the state and update the state.

        This is a three-way sync, i.e. the objects are compared to the state as base and the differences are resolved.
        """
        _logger.debug('Performing state sync...')
        for notion_id, state_obj_dct in state.objs.items():
            notion_obj, other_obj = notion_objs[notion_id], other_objs[state.ids[notion_id]]
            notion_obj_dct, other_obj_dct = self.notion_to_dict(notion_obj), self.other_to_dict(other_obj)

            for notion_attr, other_attr in self.attr_map.items():
                if notion_obj_dct[notion_attr] != state_obj_dct[notion_attr] == other_obj_dct[other_attr]:
                    _logger.debug(f'Updating other object with attribute {other_attr} to {notion_obj_dct[notion_attr]}')
                    self.other_update_obj(other_obj, other_attr, notion_obj_dct[notion_attr])
                    state_obj_dct[notion_attr] = notion_obj_dct[notion_attr]
                elif notion_obj_dct[notion_attr] == state_obj_dct[notion_attr] != other_obj_dct[other_attr]:
                    _logger.debug(f'Updating Notion object with attribute {notion_attr} to {other_obj_dct[other_attr]}')
                    self.notion_update_obj(notion_obj, notion_attr, other_obj_dct[other_attr])
                    state_obj_dct[notion_attr] = notion_obj_dct[notion_attr]
                elif notion_obj_dct[notion_attr] != state_obj_dct[notion_attr] != other_obj_dct[other_attr]:
                    state_obj_dct[notion_attr] = self.resolve_conflict(notion_obj, other_obj, notion_attr, other_attr)

        return state

    def sync(self, state: State | None) -> State:
        """The actual sync operation.

        The state holds the synced objects and their attributes as a dictionary of Notion attributes.
        """
        notion_objs = {self.notion_id(obj): obj for obj in self.get_notion_objects()}
        other_objs = {self.other_id(obj): obj for obj in self.get_other_objects()}

        if state is None:
            state = self.initial_sync(notion_objs, other_objs)
        else:
            state = self.sync_notion_deleted(state, notion_objs, other_objs)
            state = self.sync_other_deleted(state, notion_objs, other_objs)
            state = self.sync_state_changes(state, notion_objs, other_objs)

        state = self.sync_notion_created(state, notion_objs)
        state = self.sync_other_created(state, other_objs)

        return state

    def __await__(self):
        """Delegate the await to the __call__ method"""
        return self().__await__()

    async def __call__(self):
        """Run the task as scheduled."""
        self.state_path.parent.mkdir(parents=True, exist_ok=True)
        state = pickle.loads(self.state_path.read_bytes()) if self.state_path.exists() else None  # noqa: S301

        while True:
            state = self.sync(state)
            self.state_path.write_bytes(pickle.dumps(state))

            if self._in_total_times is not None:
                self._in_total_times -= 1
                if self._in_total_times <= 0:
                    break

            if self._run_every_secs is not None:
                await asyncio.sleep(self._run_every_secs)

            if self._run_every_secs is None and self._in_total_times is None:
                break


all_tasks: set[SyncTask] = set()
"""All tasks that have been created so far."""


def run_all_tasks(*, debug: bool | None = None):
    """Run all scheduled tasks."""

    async def gather_all_tasks():
        return await asyncio.gather(*all_tasks)

    _logger.info('Running all scheduled tasks...')
    asyncio.run(gather_all_tasks(), debug=debug)



"""Low-level object-oriented API for Notion.

This subpackage provides an object-oriented API for Notion by wrapping the JSON-structured
requests and responses of [notion-sdk-py] into Python objects using pydantic. Also paginated
result lists by the API are resolved automatically.

The code was taken originally from [Notional] by Jason Heddings and is MIT-licensed.
Due to the Pydantic v2 migration and several other design changes, it was heavily refactored since then.

[Notional]: https://github.com/jheddings/notional
[notion-sdk-py]: https://github.com/ramnes/notion-sdk-py/
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

import httpx
import notion_client

from ultimate_notion.config import get_cfg_file
from ultimate_notion.obj_api.endpoints import NotionAPI

if TYPE_CHECKING:
    from ultimate_notion.config import Config


_logger = logging.getLogger(__name__)


def create_notion_client(cfg: Config, **kwargs: Any) -> notion_client.Client:
    """Create a Notion client with the given authentication token."""

    class LoggingClient(httpx.Client):
        """A client that logs the request and response."""

        def send(self, request: httpx.Request, **kwargs: Any) -> httpx.Response:
            msg = f'Request: {request.method} {request.url}'
            if request.content:
                msg += f'\n{request.content.decode("utf-8") if isinstance(request.content, bytes) else request.content}'
            _logger.debug(msg)

            response = super().send(request, **kwargs)

            msg = f'Response: {response.status_code} {request.url}'
            response.read()  # Ensure that the response content is fully loaded. Memory schouldn't be an issue here.
            if response.content:
                msg += (
                    f'\n{response.content.decode("utf-8") if isinstance(response.content, bytes) else response.content}'
                )
            _logger.debug(msg)

            return response

    if (auth := cfg.ultimate_notion.token) is None:
        msg = f'No Notion token found! Check {get_cfg_file()}.'
        raise RuntimeError(msg)

    # Same sane default as notion_client defines its own logger
    kwargs.setdefault('logger', logging.getLogger('notion_client'))
    kwargs.setdefault('log_level', logging.NOTSET)
    return notion_client.Client(auth=auth, client=LoggingClient(), **kwargs)


__all__ = ['NotionAPI', 'create_notion_client']

# ToDo: Recheck every model if `= None` really is needed. Maybe come up with an even smarter way
#       to differentiate between a model for sending and receiving data.
#       Idea: Use a sentinel value, e.g. API_RESPONSE = object() as default value for fields that are only given
#       by the API and not sent to the API. This way we can differentiate between the two cases.



"""Functions and classes for working with users in Notion."""

from __future__ import annotations

from typing import cast

from ultimate_notion.core import Wrapper, get_repr
from ultimate_notion.obj_api import objects as objs


class User(Wrapper[objs.User], wraps=objs.User):
    """User object for persons, bots and unknown users.

    Unknown users are users, which no longer participate in the workspace
    or were revoked access. They are represented by their ID and have
    the name `Unknown User`.
    """

    @classmethod
    def wrap_obj_ref(cls, obj_ref: objs.User) -> User:
        self = cast(User, cls.__new__(cls))
        self.obj_ref = obj_ref
        return self

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        return get_repr(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, User):
            return NotImplemented
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def id(self):
        return self.obj_ref.id

    @property
    def name(self):
        return self.obj_ref.name

    @property
    def is_person(self) -> bool:
        return isinstance(self.obj_ref, objs.Person)

    @property
    def is_bot(self) -> bool:
        return isinstance(self.obj_ref, objs.Bot)

    @property
    def is_unknown(self) -> bool:
        return isinstance(self.obj_ref, objs.UnknownUser)

    @property
    def avatar_url(self):
        return self.obj_ref.avatar_url

    @property
    def email(self) -> str | None:
        if isinstance(self.obj_ref, objs.Person):
            return self.obj_ref.person.email
        else:  # it's a bot or unknown without an e-mail
            return None



"""Wrapper for various Notion API objects like parents, mentions, emojis & users.

Similar to other records, these object provide access to the primitive data structure
used in the Notion API as well as higher-level methods.

For validation the Pydantic model fields specify if a field is optional or not.
Some fields are always set, e.g. `id`, when retrieving an object but must not be set
when sending the object to the Notion API in order to create the object.
To model this behavior, the default value `None` is used for those objects, e.g.
```
class SelectOption(GenericObject)
    id: str = None  # type: ignore  # to make sure mypy doesn't complain
```
Also be aware that this is import when updating to differentiate actual set values
from default/unset values.
"""

from __future__ import annotations

import datetime as dt
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, Literal
from uuid import UUID

import pendulum as pnd
from pydantic import Field, SerializeAsAny

from ultimate_notion.obj_api.core import GenericObject, NotionEntity, NotionObject, TypedObject, extract_id
from ultimate_notion.obj_api.enums import BGColor, Color
from ultimate_notion.utils import parse_dt_str

if TYPE_CHECKING:
    from ultimate_notion.obj_api.blocks import Block, Database, Page


class SelectOption(GenericObject):
    """Options for select & multi-select objects."""

    name: str
    id: str = None  # type: ignore  # According to docs: "These are sometimes, but not always, UUIDs."
    color: Color = Color.DEFAULT
    description: list[RichTextBaseObject] | None = None  # ToDo: Undocumented in the Notion API

    @classmethod
    def build(cls, name, color=Color.DEFAULT) -> SelectOption:
        """Create a `SelectOption` object from the given name and color."""
        return cls.model_construct(name=name, color=color)


class SelectGroup(GenericObject):
    """Group of options for status objects."""

    name: str
    id: str = None  # type: ignore  # According to docs: "These are sometimes, but not always, UUIDs."
    color: Color = Color.DEFAULT
    option_ids: list[str] = Field(default_factory=list)


class MentionMixin(ABC):
    """Mixin for objects that can be mentioned in Notion.

    This mixin adds a `mention` property to the object, which can be used to
    reference the object in a mention.
    """

    @abstractmethod
    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        """Return a mention object for this object."""


class DateRange(GenericObject, MentionMixin):
    """A Notion date range, with an optional end date."""

    start: dt.date | dt.datetime
    end: dt.date | dt.datetime | None = None
    time_zone: str | None = None

    @classmethod
    def build(cls, dt_spec: str | dt.datetime | dt.date | pnd.Interval) -> DateRange:
        """Compose a DateRange object from the given properties."""

        if isinstance(dt_spec, str):
            dt_spec = parse_dt_str(dt_spec)

        match dt_spec:
            case pnd.Interval():
                return cls.model_construct(
                    time_zone=dt_spec.start.timezone_name if isinstance(dt_spec.start, pnd.DateTime) else None,
                    start=dt_spec.start,
                    end=dt_spec.end,
                )

            case pnd.DateTime():
                return cls.model_construct(time_zone=dt_spec.timezone_name, start=dt_spec.naive(), end=None)

            case dt.datetime():
                # We don't trust the timezone of naive datetime and convert to UTC
                return cls.model_construct(time_zone='UTC', start=dt_spec.astimezone(dt.timezone.utc), end=None)

            case pnd.Date() | dt.date():
                return cls.model_construct(time_zone=None, start=dt_spec, end=None)

            case _:
                msg = f"Unsupported type for 'dt_spec': {type(dt_spec)}"
                raise TypeError(msg)

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionDate.build(self, style=style)

    def to_pendulum(self) -> pnd.DateTime | pnd.Date | pnd.Interval:
        """Convert the DateRange to a pendulum object."""
        if self.time_zone is None:
            if self.end is None:
                return pnd.instance(self.start)
            else:
                return pnd.Interval(start=pnd.instance(self.start), end=pnd.instance(self.end))
        else:
            pnd_start = pnd.instance(self.start)
            pnd_start = pnd_start.in_tz(self.time_zone) if isinstance(pnd_start, pnd.DateTime) else pnd_start
            if self.end is None:
                return pnd_start
            else:
                pnd_end = pnd.instance(self.end)
                pnd_end = pnd_end.in_tz(self.time_zone) if isinstance(pnd_end, pnd.DateTime) else pnd_end
                return pnd.Interval(start=pnd_start, end=pnd_end)

    def __str__(self) -> str:
        # ToDo: Implement the possibility to configure date format globally, maybe in the config?
        date_str = str(self.start) if self.end is None else f'{self.start} → {self.end}'
        return f'@{date_str}'  # we add the @ as Notion does it too


class LinkObject(GenericObject):
    """Reference a URL."""

    type: str = 'url'
    url: str = None  # type: ignore


class ObjectRef(GenericObject):
    """A general-purpose object reference in the Notion API."""

    id: UUID

    @classmethod
    def build(cls, ref: ParentRef | GenericObject | UUID | str) -> ObjectRef:
        """Compose a reference to an object from the given reference.

        Strings may be either UUID's or URL's to Notion content.
        """

        match ref:
            case ObjectRef():
                return ref.model_copy(deep=True)
            case ParentRef():
                # ParentRefs are typed objects with a nested UUID
                return cls.model_construct(id=ref.value)
            case GenericObject() if hasattr(ref, 'id'):
                # Re-compose the ObjectReference from the internal ID
                return cls.build(ref.id)
            case UUID():
                return cls.model_construct(id=ref)
            case str() if (id_str := extract_id(ref)) is not None:
                return cls.model_construct(id=UUID(id_str))
            case _:
                msg = f'Cannot interpret {ref} of type {type(ref)} as a reference to an object.'
                raise ValueError(msg)


def get_uuid(obj: str | UUID | ParentRef | NotionObject | BlockRef) -> UUID:
    """Retrieves a UUID from an object reference.

    Only meant for internal use.
    """
    return ObjectRef.build(obj).id


class ParentRef(TypedObject, ABC, polymorphic_base=True):
    """Reference another block as a parent.

    Notion API: [Parent Object](https://developers.notion.com/reference/parent-object)

    Note: This class is simply a placeholder for the typed concrete *Ref classes.
          Callers should always instantiate the intended concrete versions.
    """


class DatabaseRef(ParentRef, type='database_id'):
    """Reference a database."""

    database_id: UUID

    @classmethod
    def build(cls, db_ref: Database | str | UUID) -> DatabaseRef:
        """Compose a DatabaseRef from the given reference object."""
        ref = ObjectRef.build(db_ref)
        return DatabaseRef.model_construct(database_id=ref.id)


class PageRef(ParentRef, type='page_id'):
    """Reference a page."""

    page_id: UUID

    @classmethod
    def build(cls, page_ref: Page | str | UUID) -> PageRef:
        """Compose a PageRef from the given reference object."""
        ref = ObjectRef.build(page_ref)
        return PageRef.model_construct(page_id=ref.id)


class BlockRef(ParentRef, type='block_id'):
    """Reference a block."""

    block_id: UUID

    @classmethod
    def build(cls, block_ref: Block | str | UUID) -> BlockRef:
        """Compose a BlockRef from the given reference object."""
        ref = ObjectRef.build(block_ref)
        return BlockRef.model_construct(block_id=ref.id)


class WorkspaceRef(ParentRef, type='workspace'):
    """Reference the workspace."""

    workspace: bool = True


class CommentRef(ParentRef, type='comment_id'):
    """Reference a comment."""

    comment_id: UUID

    @classmethod
    def build(cls, comment_ref: Comment | str | UUID) -> CommentRef:
        """Compose a CommentRef from the given reference object."""
        ref = ObjectRef.build(comment_ref)
        return CommentRef.model_construct(comment_id=ref.id)


class UserRef(NotionObject, object='user'):
    """Reference to a user, e.g. in `created_by`, `last_edited_by`, mentioning, etc."""

    @classmethod
    def build(cls, user_ref: User | str | UUID) -> UserRef:
        """Compose a PageRef from the given reference object."""
        ref = ObjectRef.build(user_ref)
        return UserRef.model_construct(id=ref.id)


class User(UserRef, TypedObject, MentionMixin, polymorphic_base=True):
    """Represents a User in Notion."""

    id: UUID = None  # type: ignore
    name: str | None = None
    avatar_url: str | None = None

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionUser.build(self, style=style)


class Person(User, type='person'):
    """Represents a Person in Notion."""

    class TypeData(GenericObject):
        email: str = None  # type: ignore

    person: TypeData = TypeData()


class Bot(User, type='bot'):
    """Represents a Bot in Notion."""

    class TypeData(GenericObject):
        owner: WorkspaceRef = None  # type: ignore
        workspace_name: str = None  # type: ignore

    bot: TypeData = TypeData()


class UnknownUser(User, type='unknown'):
    """Represents an unknown user in Notion.

    This is a unofficial placeholder for a user that is not recognized by the API.
    """

    name: Literal['Unknown User'] = 'Unknown User'

    class TypeData(GenericObject): ...

    unknown: TypeData = TypeData()


class EmojiObject(TypedObject, type='emoji'):
    """A Notion emoji object."""

    emoji: str

    @classmethod
    def build(cls, emoji: str) -> EmojiObject:
        """Compose an EmojiObject from the given emoji string."""
        return EmojiObject.model_construct(emoji=emoji)


class Annotations(GenericObject):
    """Style information for RichTextObject's."""

    bold: bool = False
    italic: bool = False
    strikethrough: bool = False
    underline: bool = False
    code: bool = False
    color: Color | BGColor = Color.DEFAULT


class RichTextBaseObject(TypedObject, polymorphic_base=True):
    """Base class for Notion rich text elements."""

    plain_text: str
    href: str | None = None
    annotations: Annotations | None = None


def rich_text_to_str(rich_texts: list[RichTextBaseObject]) -> str:
    """Convert a list of rich texts to plain text."""
    return ''.join(rich_text.plain_text for rich_text in rich_texts)


class TextObject(RichTextBaseObject, type='text'):
    """Notion text element."""

    class TypeData(GenericObject):
        content: str = None  # type: ignore
        link: LinkObject | None = None

    text: TypeData = TypeData()

    @classmethod
    def build(cls, text: str, *, href: str | None = None, style: Annotations | None = None) -> TextObject:
        """Compose a TextObject from the given properties.

        Args:
            text: the plain text of this object
            href: optional link for this object
            style: optional annotations for styling this text
        """
        link = LinkObject(url=href) if href else None
        nested = cls.TypeData(content=text, link=link)
        style = deepcopy(style)

        return cls.model_construct(
            plain_text=text,
            text=nested,
            href=href,
            annotations=style,
        )


class EquationObject(RichTextBaseObject, type='equation'):
    """Notion equation element."""

    class TypeData(GenericObject):
        expression: str

    equation: TypeData

    @classmethod
    def build(cls, expression: str, *, href: str | None = None, style: Annotations | None = None) -> EquationObject:
        """Compose a TextObject from the given properties.

        Args:
            expression: expression
            href: optional link for this object
            style: optional annotations for styling this text
        """
        style = deepcopy(style)
        return cls.model_construct(
            plain_text=expression, equation=cls.TypeData(expression=expression), href=href, annotations=style
        )


class MentionBase(TypedObject, polymorphic_base=True):
    """Base class for typed `Mention` objects."""


class MentionObject(RichTextBaseObject, type='mention'):
    """Notion mention element."""

    mention: SerializeAsAny[MentionBase]


class MentionUser(MentionBase, type='user'):
    """Nested user data for `Mention` properties."""

    user: SerializeAsAny[User]

    @classmethod
    def build(cls, user: User, *, style: Annotations | None = None) -> MentionObject:
        style = deepcopy(style)
        mention = cls.model_construct(user=user)
        # note that `href` is always `None` for user mentions, also we prepend the '@' to mimic server side
        return MentionObject.model_construct(plain_text=f'@{user.name}', href=None, annotations=style, mention=mention)


class MentionPage(MentionBase, type='page'):
    """Nested page data for `Mention` properties."""

    page: SerializeAsAny[ObjectRef]

    @classmethod
    def build(cls, page: Page, *, style: Annotations | None = None) -> MentionObject:
        style = deepcopy(style)
        page_ref = ObjectRef.build(page)
        mention = cls.model_construct(page=page_ref)
        # note that `href` is always `None` for page mentions
        return MentionObject.model_construct(
            plain_text=rich_text_to_str(page.title), href=None, annotations=style, mention=mention
        )


class MentionDatabase(MentionBase, type='database'):
    """Nested database information for `Mention` properties."""

    database: SerializeAsAny[ObjectRef]

    @classmethod
    def build(cls, db: Database, *, style: Annotations | None = None) -> MentionObject:
        style = deepcopy(style)
        db_ref = ObjectRef.build(db)
        mention = cls.model_construct(database=db_ref)
        # note that `href` is always `None` for database mentions
        return MentionObject.model_construct(
            plain_text=rich_text_to_str(db.title), ref=None, annotations=style, mention=mention
        )


class MentionDate(MentionBase, type='date'):
    """Nested date data for `Mention` properties."""

    date: DateRange

    @classmethod
    def build(cls, date_range: DateRange, *, style: Annotations | None = None) -> MentionObject:
        style = deepcopy(style)
        mention = cls.model_construct(date=date_range)
        # note that `href` is always `None` for date mentions
        return MentionObject.model_construct(plain_text=str(date_range), ref=None, annotations=style, mention=mention)


class MentionLinkPreview(MentionBase, type='link_preview'):
    """Nested url data for `Mention` properties.

    !!! warning

        Link previews cannot be created via the API.
    """

    class TypeData(GenericObject):
        url: str

    link_preview: TypeData


class MentionTemplateData(TypedObject):
    """Nested template data for `Mention` properties."""


class MentionTemplate(MentionBase, type='template_mention'):
    """Nested template data for `Mention` properties."""

    template_mention: SerializeAsAny[MentionTemplateData]


class MentionTemplateDate(MentionTemplateData, type='template_mention_date'):
    """Nested date template data for `Mention` properties."""

    template_mention_date: str


class MentionTemplateUser(MentionTemplateData, type='template_mention_user'):
    """Nested user template data for `Mention` properties."""

    template_mention_user: str


class FileObject(TypedObject, polymorphic_base=True):
    """A Notion file object.

    Depending on the context, a FileObject may require a name (such as in the `Files`
    property). This makes the object hierarchy difficult, so here we simply allow
    `name` to be optional. It is the responsibility of the caller to set `name` if
    required by the API.
    """

    name: str | None = None
    caption: list[SerializeAsAny[RichTextBaseObject]] | None = None


class HostedFile(FileObject, type='file'):
    """A Notion file object."""

    class TypeData(GenericObject):
        url: str
        expiry_time: dt.datetime | None = None

    file: TypeData

    @classmethod
    def build(
        cls,
        url: str,
        *,
        name: str | None = None,
        caption: list[RichTextBaseObject] | None = None,
        expiry_time: dt.datetime | None = None,
    ) -> HostedFile:
        """Create a new `HostedFile` from the given URL."""
        return cls.model_construct(name=name, caption=caption, file=cls.TypeData(url=url, expiry_time=expiry_time))


class ExternalFile(FileObject, type='external'):
    """A Notion external file object."""

    class TypeData(GenericObject):
        url: str

    external: TypeData

    @classmethod
    def build(
        cls, url: str, *, name: str | None = None, caption: list[RichTextBaseObject] | None = None
    ) -> ExternalFile:
        """Create a new `ExternalFile` from the given URL."""
        return cls.model_construct(name=name, caption=caption, external=cls.TypeData(url=url))


class Comment(NotionEntity, object='comment'):
    """A Notion comment object."""

    discussion_id: UUID
    rich_text: list[RichTextBaseObject]



"""Wrapper for Notion API blocks.

Blocks are the base for all Notion content.
"""

from __future__ import annotations

from abc import ABC
from typing import Any, cast
from uuid import UUID

from pydantic import Field, SerializeAsAny

from ultimate_notion.obj_api.core import GenericObject, NotionEntity, TypedObject
from ultimate_notion.obj_api.enums import BGColor, CodeLang, Color
from ultimate_notion.obj_api.objects import (
    Annotations,
    BlockRef,
    EmojiObject,
    FileObject,
    MentionDatabase,
    MentionMixin,
    MentionObject,
    MentionPage,
    ParentRef,
    RichTextBaseObject,
    UserRef,
)
from ultimate_notion.obj_api.props import PropertyValue, Title
from ultimate_notion.obj_api.schema import PropertyType


class DataObject(NotionEntity, ABC):
    """The base type for all Notion objects that hold actual data."""

    has_children: bool = False

    in_trash: bool = False  # used to be `archived`
    archived: bool = False  # ToDo: Deprecated but still partially used in Notion. Check to remove in v1.0!

    last_edited_by: UserRef = None  # type: ignore


class Database(DataObject, MentionMixin, object='database'):
    """A database record type."""

    title: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
    url: str = None  # type: ignore
    public_url: str | None = None
    icon: SerializeAsAny[FileObject] | EmojiObject | None = None
    cover: SerializeAsAny[FileObject] | None = None
    properties: dict[str, PropertyType] = None  # type: ignore
    description: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
    is_inline: bool = False

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionDatabase.build(self, style=style)


class Page(DataObject, MentionMixin, object='page'):
    """A standard Notion page object."""

    url: str = None  # type: ignore
    public_url: str | None = None
    icon: SerializeAsAny[FileObject] | EmojiObject | None = None
    cover: SerializeAsAny[FileObject] | None = None
    properties: dict[str, PropertyValue] = None  # type: ignore

    def _get_title_prop_name(self) -> str:
        """Get the name of the title property."""
        # As the 'title' property might be renamed in case of pages in databases, we look for the `id`.
        for name, prop in self.properties.items():
            if prop.id == 'title':
                return name
        msg = 'Encountered a page without title property'
        raise RuntimeError(msg)

    @property
    def title(self) -> list[RichTextBaseObject]:
        """Retrieve the title of the page from page properties."""
        title_prop_name = self._get_title_prop_name()
        title_prop = cast(Title, self.properties[title_prop_name])
        return title_prop.title

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionPage.build(self, style=style)


class Block(DataObject, TypedObject, object='block', polymorphic_base=True):
    """A standard block object in Notion.

    Calling the block will expose the nested data in the object.
    """

    id: UUID = None  # type: ignore


class UnsupportedBlock(Block, type='unsupported'):
    """A placeholder for unsupported blocks in the API."""

    class TypeData(GenericObject): ...

    unsupported: TypeData | None = None


class TextBlock(Block, ABC):
    """A standard abstract text block object in Notion."""


class Paragraph(TextBlock, type='paragraph'):
    """A paragraph block in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)
        color: Color | BGColor = Color.DEFAULT

    paragraph: TypeData = TypeData()


class Heading(TextBlock, ABC):
    """Abstract Heading block."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        color: Color | BGColor = Color.DEFAULT
        is_toggleable: bool = False


class Heading1(Heading, type='heading_1'):
    """A heading_1 block in Notion."""

    heading_1: Heading.TypeData = Heading.TypeData()


class Heading2(Heading, type='heading_2'):
    """A heading_2 block in Notion."""

    heading_2: Heading.TypeData = Heading.TypeData()


class Heading3(Heading, type='heading_3'):
    """A heading_3 block in Notion."""

    heading_3: Heading.TypeData = Heading.TypeData()


class Quote(TextBlock, type='quote'):
    """A quote block in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)
        color: Color | Color = Color.DEFAULT

    quote: TypeData = TypeData()


class Code(TextBlock, type='code'):
    """A code block in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        caption: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        language: CodeLang = CodeLang.PLAIN_TEXT

    code: TypeData = TypeData()


class Callout(TextBlock, type='callout'):
    """A callout block in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)
        icon: SerializeAsAny[FileObject] | EmojiObject = None  # type: ignore
        color: Color | BGColor = BGColor.GRAY

    callout: TypeData = TypeData()


class BulletedListItem(TextBlock, type='bulleted_list_item'):
    """A bulleted list item in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)
        color: Color | BGColor = Color.DEFAULT

    bulleted_list_item: TypeData = TypeData()


class NumberedListItem(TextBlock, type='numbered_list_item'):
    """A numbered list item in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)
        color: Color | BGColor = Color.DEFAULT

    numbered_list_item: TypeData = TypeData()


class ToDo(TextBlock, type='to_do'):
    """A todo list item in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        checked: bool = False
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)
        color: Color | BGColor = Color.DEFAULT

    to_do: TypeData = TypeData()


class Toggle(TextBlock, type='toggle'):
    """A toggle list item in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] = None  # type: ignore
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)
        color: Color | BGColor = Color.DEFAULT

    toggle: TypeData = TypeData()


class Divider(Block, type='divider'):
    """A divider block in Notion."""

    class TypeData(GenericObject): ...

    divider: TypeData = TypeData()


class TableOfContents(Block, type='table_of_contents'):
    """A table_of_contents block in Notion."""

    class TypeData(GenericObject):
        color: Color | BGColor = Color.DEFAULT

    table_of_contents: TypeData = TypeData()


class Breadcrumb(Block, type='breadcrumb'):
    """A breadcrumb block in Notion."""

    class TypeData(GenericObject): ...

    breadcrumb: TypeData = TypeData()


class Embed(Block, type='embed'):
    """An embed block in Notion."""

    class TypeData(GenericObject):
        url: str = None  # type: ignore
        caption: list[SerializeAsAny[RichTextBaseObject]] | None = None

    embed: TypeData = TypeData()


class Bookmark(Block, type='bookmark'):
    """A bookmark block in Notion."""

    class TypeData(GenericObject):
        url: str = None  # type: ignore
        caption: list[SerializeAsAny[RichTextBaseObject]] | None = None

    bookmark: TypeData = TypeData()


class LinkPreview(Block, type='link_preview'):
    """A link_preview block in Notion."""

    class TypeData(GenericObject):
        url: str = None  # type: ignore

    link_preview: TypeData = TypeData()


class Equation(Block, type='equation'):
    """An equation block in Notion."""

    class TypeData(GenericObject):
        expression: str | None = None

    equation: TypeData = TypeData()


class FileBase(Block, ABC):
    """A abstract block referencing a FileObject."""


class File(FileBase, type='file'):
    """A file block in Notion."""

    file: SerializeAsAny[FileObject] = None  # type: ignore


class Image(FileBase, type='image'):
    """An image block in Notion."""

    image: SerializeAsAny[FileObject] = None  # type: ignore


class Video(FileBase, type='video'):
    """A video block in Notion."""

    video: SerializeAsAny[FileObject] = None  # type: ignore


class PDF(FileBase, type='pdf'):
    """A pdf block in Notion."""

    pdf: SerializeAsAny[FileObject] = None  # type: ignore


class ChildPage(Block, type='child_page'):
    """A child page block in Notion."""

    class TypeData(GenericObject):
        title: str = None  # type: ignore

    child_page: TypeData = TypeData()


class ChildDatabase(Block, type='child_database'):
    """A child database block in Notion."""

    class TypeData(GenericObject):
        title: str = None  # type: ignore

    child_database: TypeData = TypeData()


class Column(Block, type='column'):
    """A column block in Notion."""

    class TypeData(GenericObject):
        # note that children will not be populated when getting this block
        # https://developers.notion.com/changelog/column-list-and-column-support
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)

    column: TypeData = TypeData()

    @classmethod
    def build(cls) -> Column:
        return Column.model_construct(column=cls.TypeData.model_construct(children=[]))


class ColumnList(Block, type='column_list'):
    """A column list block in Notion."""

    class TypeData(GenericObject):
        # note that children will not be populated when getting this block
        # https://developers.notion.com/changelog/column-list-and-column-support
        children: list[Column] = Field(default_factory=list)

    column_list: TypeData = TypeData()


class TableRow(Block, type='table_row'):
    """A table_row block in Notion."""

    class TypeData(GenericObject):
        cells: list[list[SerializeAsAny[RichTextBaseObject]]] = Field(default_factory=list)

    table_row: TypeData = TypeData()

    @classmethod
    def build(cls, n_cells: int) -> TableRow:
        return TableRow.model_construct(table_row=cls.TypeData.model_construct(cells=[[] for _ in range(n_cells)]))


class Table(Block, type='table'):
    """A table block in Notion."""

    class TypeData(GenericObject):
        table_width: int = 0
        has_column_header: bool = False
        has_row_header: bool = False

        # note that children will not be populated when getting this block
        # https://developers.notion.com/reference/block#table-blocks
        children: list[TableRow] = Field(default_factory=list)

    table: TypeData = TypeData()


class LinkToPage(Block, type='link_to_page'):
    """A link_to_page block in Notion."""

    link_to_page: SerializeAsAny[ParentRef] = None  # type: ignore


class SyncedBlock(Block, type='synced_block'):
    """A synced_block block in Notion - either original or synced."""

    class TypeData(GenericObject):
        synced_from: BlockRef | None = None
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)

    synced_block: TypeData = TypeData()

    def serialize_for_api(self) -> dict[str, Any]:
        """Serialize the object for sending it to the Notion API."""
        model_data = super().serialize_for_api()
        # Everywhere else we have to remove "null" values but `synced_from` is an exception
        if self.synced_block.synced_from is None:
            model_data['synced_block']['synced_from'] = None
        return model_data


class Template(Block, type='template'):
    """A template block in Notion."""

    class TypeData(GenericObject):
        rich_text: list[SerializeAsAny[RichTextBaseObject]] | None = None
        children: list[SerializeAsAny[Block]] = Field(default_factory=list)

    template: TypeData = TypeData()



"""Functionality around defining a database schema.

Currently only normal databases, no wiki databases, can be created [1].
Neither the `Unique ID` nor `Status` nor the `Verfication` page property can be set as a database property
in a custom Schema when creating the database.

[1] https://developers.notion.com/docs/working-with-databases#wiki-databases


### Design Principles

A schema is a subclass of `Schema` that holds `Property` objects with a name and an
actual `PropertyType`, e.g. `Text`, `Number`.

The source of truth is always the `obj_ref` and a `PropertyType` holds only auxilliary
information if actually needed. Since the object references `obj_ref` must always point
to the actual `obj_api.blocks.Database.properties` value if the schema is bound to an database,
the method `_remap_obj_refs` rewires this when a schema is used to create a database.
"""

from __future__ import annotations

from abc import ABCMeta
from collections.abc import Iterator
from textwrap import dedent
from typing import TYPE_CHECKING, Any, TypeVar, cast

from tabulate import tabulate

import ultimate_notion.obj_api.schema as obj_schema
from ultimate_notion import rich_text
from ultimate_notion.core import Wrapper, get_active_session, get_repr
from ultimate_notion.obj_api.schema import AggFunc, NumberFormat
from ultimate_notion.option import Option, OptionGroup, OptionNS
from ultimate_notion.props import PropertyValue
from ultimate_notion.utils import EmptyListError, SList, dict_diff_str, is_notebook

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.page import Page


class SchemaError(Exception):
    """Raised when there are issues with the schema of a database."""

    def __init__(self, message):
        """Initialize the `SchemaError` with a supplied message."""
        super().__init__(message)


class SchemaNotBoundError(SchemaError):
    """Raised when the schema is not bound to a database."""

    def __init__(self, schema: type[Schema]):
        self.schema = schema
        msg = f'Schema {schema.__name__} is not bound to any database'
        super().__init__(msg)


class ReadOnlyPropertyError(SchemaError):
    """Raised when an attempt is made to write to a write-protected property."""

    def __init__(self, prop: Property):
        self.prop = prop
        msg = f"Argument {prop.attr_name} refers to the read-only propert '{prop.name}' of type {prop.type}"
        super().__init__(msg)


T = TypeVar('T', bound=obj_schema.PropertyType)


class PropertyType(Wrapper[T], wraps=obj_schema.PropertyType):
    """Base class for Notion property objects.

    Property types define the value types of properties in a database, e.g. number, date, text, etc.
    """

    obj_ref: T
    """Reference to the low-level object representation of this property type"""
    allowed_at_creation = True
    """If the Notion API allows to create a new database with a property of this type"""
    prop_ref: Property | None = None
    """Back reference to the property having this type"""

    @property
    def id(self) -> str | None:
        """Return identifier of this property type."""
        return self.obj_ref.id

    @property
    def name(self) -> str | None:
        """Return name of this property type."""
        return self.obj_ref.name

    @property
    def prop_value(self) -> type[PropertyValue]:
        """Return the corresponding property value of this property type."""
        return PropertyValue._type_value_map[self.obj_ref.type]

    @property
    def readonly(self) -> bool:
        """Return if this property type is read-only."""
        return self.prop_value.readonly

    @property
    def _is_init(self) -> bool:
        """Determines if the property type is already initialized"""
        return hasattr(self, 'obj_ref')

    def __init__(self, *args, **kwargs):
        obj_api_type = self._obj_api_map_inv[self.__class__]
        self.obj_ref = obj_api_type.build(*args, **kwargs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PropertyType):
            return NotImplemented
        return self.obj_ref.type == other.obj_ref.type and self.obj_ref.value == self.obj_ref.value

    def __hash__(self) -> int:
        return hash(self.obj_ref.type) + hash(self.obj_ref.value)

    def __repr__(self) -> str:
        return get_repr(self, name='PropertyType', desc=self.__class__.__name__)

    def __str__(self) -> str:
        return self.__class__.__name__


class Property:
    """Database property/column with a name and a certain property type for defining a Notion database schema.

    This is implemented as a descriptor.
    """

    _name: str
    _type: PropertyType
    # properties below are set by __set_name__
    _schema: type[Schema]
    _attr_name: str  # Python attribute name of the property in the schema

    def __init__(self, name: str, type: PropertyType) -> None:  # noqa: A002
        self._name = name
        self._type = type

    def __set_name__(self, owner: type[Schema], name: str):
        self._schema = owner
        self._attr_name = name
        self._type.prop_ref = self  # link back to allow access to _schema, _py_name e.g. for relations

    def __repr__(self) -> str:
        return get_repr(self, name='Property', desc=self.type)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str):
        raise NotImplementedError

    @property
    def type(self) -> PropertyType:
        return self._type

    @type.setter
    def type(self, new_type: PropertyType):
        raise NotImplementedError

    @property
    def attr_name(self) -> str:
        return self._attr_name


class SchemaRepr(ABCMeta):
    """Metaclass for the schema of a database.

    This makes the schema class itself more user-friendly by providing a custom `__repr__` method
    and letting it behave like a dictionary for the properties athough it is a class, not an instance.
    """

    def __repr__(cls) -> str:
        # We can only overwrite __repr__ for a class in a metaclass
        if not issubclass(cls, Schema):
            msg = 'Metaclass SchemaRepr can only be used with subclasses of Schema'
            raise TypeError(msg)
        return cls.as_table(tablefmt='simple')

    def __getitem__(cls, prop_name: str) -> PropertyType:
        if not issubclass(cls, Schema):
            msg = 'Metaclass SchemaRepr can only be used with subclasses of Schema'
            raise TypeError(msg)
        return cls.get_prop(prop_name).type

    def __len__(cls) -> int:
        if not issubclass(cls, Schema):
            msg = 'Metaclass SchemaRepr can only be used with subclasses of Schema'
            raise TypeError(msg)
        return len(cls.get_props())

    def __iter__(cls) -> Iterator[PropertyType]:
        if not issubclass(cls, Schema):
            msg = 'Metaclass SchemaRepr can only be used with subclasses of Schema'
            raise TypeError(msg)
        return (prop.type for prop in cls.get_props())


class Schema(metaclass=SchemaRepr):
    """Base class for the schema of a database."""

    db_title: rich_text.Text | None
    db_desc: rich_text.Text | None
    _database: Database | None = None

    def __init_subclass__(cls, db_title: str | None, **kwargs: Any):
        if db_title is not None:
            db_title = rich_text.Text(db_title)
        cls.db_title = db_title

        cls.db_desc = rich_text.Text(cls.__doc__) if cls.__doc__ is not None else None
        super().__init_subclass__(**kwargs)

    @classmethod
    def from_dict(
        cls, schema_dct: dict[str, PropertyType], db_title: str | None = None, db_desc: str | None = None
    ) -> type[Schema]:
        """Creation of a schema from a dictionary for easy support of dynamically created schemas."""
        title_props = [k for k, v in schema_dct.items() if isinstance(v, Title)]
        if not title_props:
            msg = 'Missing an item with property type `Title` as value'
            raise SchemaError(msg)
        elif len(title_props) > 1:
            msg = f'More than one item with property type `Title` as value found: {", ".join(title_props)}'
            raise SchemaError(msg)

        cls_name = f'{cls.__name__}FromDct'
        attrs: dict[str, Any] = {'db_desc': db_desc}
        for prop_name, prop_type in schema_dct.items():
            attrs[rich_text.snake_case(prop_name)] = Property(prop_name, prop_type)
        return type(cls_name, (Schema,), attrs, db_title=db_title)

    @classmethod
    def create(cls, **kwargs) -> Page:
        """Create a page using this schema with a bound database."""
        return cls.get_db().create_page(**kwargs)

    @classmethod
    def get_props(cls) -> list[Property]:
        """Get all properties of this schema."""
        return [prop for prop in cls.__dict__.values() if isinstance(prop, Property)]

    @classmethod
    def get_prop(cls, prop_name: str) -> Property:
        """Get a specific property from this schema assuming that property names are unique."""
        try:
            return SList([prop for prop in cls.get_props() if prop.name == prop_name]).item()
        except EmptyListError as e:
            msg = f'Property `{prop_name}` not found in database `{cls._database}`.'
            raise SchemaError(msg) from e

    @classmethod
    def to_dict(cls) -> dict[str, PropertyType]:
        """Convert this schema to a dictionary of property names and corresponding types."""
        return {prop.name: prop.type for prop in cls.get_props()}

    @classmethod
    def as_table(cls, tablefmt: str | None = None) -> str:
        """Return the schema in a given string table format.

        Some table formats:

        - plain: no pseudographics
        - simple: Pandoc's simple table, i.e. only dashes to separate header from content
        - github: GitHub flavored Markdown
        - simple_grid: uses dashes & pipes to separate cells
        - html: standard html markup

        Find more table formats under: https://github.com/astanin/python-tabulate#table-format
        """
        if tablefmt is None:
            tablefmt = 'html' if is_notebook() else 'simple'

        headers = ['Name', 'Property', 'Attribute']
        rows = []
        for prop in cls.get_props():
            rows.append((prop.name, prop.type, prop.attr_name))

        return tabulate(rows, headers=headers, tablefmt=tablefmt)

    @classmethod
    def show(cls, *, simple: bool | None = None):
        """Show the schema as html or as simple table."""
        if simple:
            tablefmt = 'simple'
        elif simple is None:
            tablefmt = 'html' if is_notebook() else 'simple'
        else:
            tablefmt = 'html'

        table_str = cls.as_table(tablefmt)

        if is_notebook() and (tablefmt == 'html'):
            from IPython.display import display_html  # noqa: PLC0415

            display_html(table_str)
        else:
            print(table_str)  # noqa: T201

    @classmethod
    def _repr_html_(cls) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this schema."""
        return cls.as_table(tablefmt='html')

    @classmethod
    def get_title_prop(cls) -> Property:
        """Returns the property holding the title of the pages."""
        return SList(prop for prop in cls.get_props() if isinstance(prop.type, Title)).item()

    @classmethod
    def assert_consistency_with(cls, other_schema: type[Schema], *, during_init: bool = False) -> None:
        """Assert that this schema is consistent with another schema."""
        own_schema_dct = cls.to_dict()
        other_schema_dct = other_schema.to_dict()

        if during_init:
            # backward relation was not yet initialised in the other schema (during the creation of the data model)
            # or self-referencing relation was not yet initialised
            other_schema_dct = {
                name: prop_type
                for name, prop_type in other_schema_dct.items()
                if not (
                    (isinstance(prop_type, Relation) and not prop_type.schema)
                    or (isinstance(prop_type, Rollup) and prop_type.is_self_ref)
                )
            }

        if other_schema_dct != own_schema_dct:
            props_added, props_removed, props_changed = dict_diff_str(own_schema_dct, other_schema_dct)
            msg = f"""Provided schema is not consistent with the current schema of the database:
                      Properties added: {props_added}
                      Properties removed: {props_removed}
                      Properties changed: {props_changed}
                   """
            raise SchemaError(dedent(msg))

    @classmethod
    def get_db(cls) -> Database:
        """Get the database that is bound to this schema."""
        if cls._database is not None:  # is_bound() cannot be used here due to type checker
            return cls._database
        else:
            raise SchemaNotBoundError(cls)

    @classmethod
    def bind_db(cls, db: Database):
        """Bind the PageSchema to the corresponding database for back-reference."""
        cls._database = db
        cls._set_obj_refs()

    @classmethod
    def is_bound(cls) -> bool:
        """Determines if the schema is bound to a database."""
        return cls._database is not None

    @classmethod
    def _get_fwd_rels(cls) -> list[Property]:
        return [
            prop
            for prop in cls.get_props()
            if isinstance(prop.type, Relation) and not (prop.type._is_two_way_target or prop.type.is_self_ref)
        ]

    @classmethod
    def _init_fwd_rels(cls):
        """Initialise all non-self-referencing forward relations assuming that the target schemas exist."""
        for prop in cls._get_fwd_rels():
            prop_type = cast(Relation, prop.type)
            prop_type._make_obj_ref()

    @classmethod
    def _get_self_refs(cls) -> list[Property]:
        """Get all self-referencing relation properties."""
        return [prop for prop in cls.get_props() if isinstance(prop.type, Relation) and prop.type.is_self_ref]

    @classmethod
    def _has_self_refs(cls) -> bool:
        """Determine if self-referencing relation properties are present."""
        return bool([prop for prop in cls.get_props() if isinstance(prop.type, Relation) and prop.type.is_self_ref])

    @classmethod
    def _init_self_refs(cls):
        """Initialise all forward self-referencing relations."""
        if not cls._has_self_refs():
            return
        db = cls.get_db()  # raises if not bound!
        for prop in cls._get_self_refs():
            prop_type = cast(Relation, prop.type)
            prop_type._schema = cls  # replace placeholder `SelfRef` with this schema
            prop_type._make_obj_ref()

        new_props_schema = {prop.name: prop.type.obj_ref for prop in cls._get_self_refs()} or None
        session = get_active_session()
        db.obj_ref = session.api.databases.update(db.obj_ref, schema=new_props_schema)
        cls._set_obj_refs()

    @classmethod
    def _init_self_ref_rollups(cls):
        """Initialise all rollup properties that reference the same schema."""
        if not cls._has_self_refs():
            return
        db = cls.get_db()  # raises if not bound!

        self_ref_rollups = [prop for prop in cls.get_props() if isinstance(prop.type, Rollup) and prop.type.is_self_ref]
        new_props_schema = {prop.name: prop.type.obj_ref for prop in self_ref_rollups} or None
        session = get_active_session()
        db.obj_ref = session.api.databases.update(db.obj_ref, schema=new_props_schema)
        cls._set_obj_refs()

    @classmethod
    def _get_init_props(cls) -> list[Property]:
        """Get all properties that are initialized by now."""
        return [prop for prop in cls.get_props() if prop.type._is_init]

    @classmethod
    def _update_bwd_rels(cls):
        """Update the default property name in case of a two-way relation in the external target schema.

        By default the property in the target schema is named "Related to <this_database> (<this_field>)"
        which is then set to the name specified as backward relation.
        """
        for prop_type in cls.to_dict().values():
            if isinstance(prop_type, Relation) and prop_type.is_two_way:
                prop_type._update_bwd_rel()

    @classmethod
    def _set_obj_refs(cls):
        """Map obj_refs from the properties of the schema to obj_ref.properties of the bound database."""
        db_props_dct = cls.get_db().obj_ref.properties
        for prop_name, prop_type in cls.to_dict().items():
            obj_ref = db_props_dct.get(prop_name)
            if obj_ref:
                prop_type.obj_ref = obj_ref


class Title(PropertyType[obj_schema.Title], wraps=obj_schema.Title):
    """Defines the mandatory title property in a database."""


class Text(PropertyType[obj_schema.RichText], wraps=obj_schema.RichText):
    """Defines a text property in a database."""


class Number(PropertyType[obj_schema.Number], wraps=obj_schema.Number):
    """Defines a number property in a database."""

    def __init__(self, number_format: NumberFormat | str = NumberFormat.NUMBER):
        super().__init__(NumberFormat(number_format))


class Select(PropertyType[obj_schema.Select], wraps=obj_schema.Select):
    """Defines a select property in a database."""

    def __init__(self, options: list[Option] | type[OptionNS]):
        if isinstance(options, type) and issubclass(options, OptionNS):
            options = options.to_list()

        option_objs = [option.obj_ref for option in options]
        super().__init__(option_objs)

    @property
    def options(self) -> dict[str, Option]:
        return {option.name: Option.wrap_obj_ref(option) for option in self.obj_ref.select.options}


class MultiSelect(PropertyType[obj_schema.MultiSelect], wraps=obj_schema.MultiSelect):
    """Defines a multi-select property in a database."""

    def __init__(self, options: list[Option] | type[OptionNS]):
        if isinstance(options, type) and issubclass(options, OptionNS):
            options = options.to_list()

        option_objs = [option.obj_ref for option in options]
        super().__init__(option_objs)

    @property
    def options(self) -> dict[str, Option]:
        return {option.name: Option.wrap_obj_ref(option) for option in self.obj_ref.multi_select.options}


class Status(PropertyType[obj_schema.Status], wraps=obj_schema.Status):
    """Defines a status property in a database.

    The Notion API doesn't allow to create a property of this type.
    Sending it to the API with options and option groups defined results in an error
    about the existence of the keys `options` and `groups` and removing them
    creates a database with the property missing... ignorance is bliss.

    Also the Status configuration is not mentioned as a
    [Property Schema Object])https://developers.notion.com/reference/property-schema-object).
    """

    allowed_at_creation = False

    @property
    def options(self) -> list[Option]:
        return [Option.wrap_obj_ref(option) for option in self.obj_ref.status.options]

    @property
    def groups(self) -> list[OptionGroup]:
        return [OptionGroup.wrap_obj_ref(group, options=self.options) for group in self.obj_ref.status.groups]


class Date(PropertyType[obj_schema.Date], wraps=obj_schema.Date):
    """Defines a date property in a database."""


class Person(PropertyType[obj_schema.People], wraps=obj_schema.People):
    """Defines a person/people property in a database."""


class Files(PropertyType[obj_schema.Files], wraps=obj_schema.Files):
    """Defines a files property in a database."""


class Checkbox(PropertyType[obj_schema.Checkbox], wraps=obj_schema.Checkbox):
    """Defines a checkbox property in database."""


class Email(PropertyType[obj_schema.Email], wraps=obj_schema.Email):
    """Defines an e-mail property in a database."""


class URL(PropertyType[obj_schema.URL], wraps=obj_schema.URL):
    """Defines a URL property in a database."""


class Phone(PropertyType[obj_schema.PhoneNumber], wraps=obj_schema.PhoneNumber):
    """Defines a phone number property in a database."""


class Formula(PropertyType[obj_schema.Formula], wraps=obj_schema.Formula):
    """Defines a formula property in a database.

    Currently the formula expression cannot reference other formula properties, e.g. `prop("other formula")`
    This is a limitation of the API.
    """

    def __init__(self, expression: str):
        super().__init__(expression)

    @property
    def expression(self) -> str:
        return self.obj_ref.formula.expression


class RelationError(SchemaError):
    """Error if a Relation cannot be initialised."""


class SelfRef(Schema, db_title=None):
    """Target schema for self-referencing database relations."""


class Relation(PropertyType[obj_schema.Relation], wraps=obj_schema.Relation):
    """Relation to another database."""

    _schema: type[Schema] | None = None  # other schema, i.e. of the target database
    _two_way_prop: Property | None = None  # other property, i.e. of the target database

    def __init__(self, schema: type[Schema] | None = None, *, two_way_prop: Property | None = None):
        # Note that we don't call super().__init__ here since we only know how to build the obj_ref later.
        if two_way_prop and not schema:
            msg = '`schema` needs to be provided if `two_way_prop` is set'
            raise RuntimeError(msg)

        if isinstance(schema, Property):
            msg = 'Please provide a schema, not a property! Use `two_way_prop` to specify a property.'
            raise ValueError(msg)

        self._schema = schema

        if two_way_prop is not None:
            if not isinstance(two_way_prop.type, Relation):
                msg = f'The two-way property {two_way_prop.name} needs to be of type Relation!'
                raise ValueError(msg)

            if two_way_prop.type.schema is not None:
                msg = f'The two-way property {two_way_prop.name} must not reference a schema itself'
                raise ValueError(msg)

            self._two_way_prop = two_way_prop

    @property
    def schema(self) -> type[Schema] | None:
        """Schema of the relation database."""
        if self._schema:
            return self._schema if self._schema is not SelfRef else None
        elif self.prop_ref is not None and self.prop_ref._schema.is_bound():
            db = self.prop_ref._schema._database
            session = get_active_session()
            return session.get_db(self.obj_ref.relation.database_id).schema if db is not None else None
        else:
            return None

    @property
    def _is_two_way_target(self) -> bool:
        """Determines if relation is meant as target for a two-way relation."""
        return self._schema is None

    @property
    def two_way_prop(self) -> Property | None:
        """Return the target property object of a two-way relation."""
        if self._two_way_prop:
            return self._two_way_prop
        elif (
            hasattr(self, 'obj_ref')
            and self.schema
            and isinstance(self.obj_ref.relation, obj_schema.DualPropertyRelation)
        ):
            prop_name = self.obj_ref.relation.dual_property.synced_property_name
            return self.schema.get_prop(prop_name) if prop_name else None
        else:
            return None

    @property
    def is_two_way(self) -> bool:
        """Determine if this relation is a two-way relation."""
        return self.two_way_prop is not None

    @property
    def is_self_ref(self) -> bool:
        """Determines if this relation is self referencing the same schema."""
        return (self._schema is SelfRef) or (self.prop_ref is not None and self._schema is self.prop_ref._schema)

    def _make_obj_ref(self) -> None:
        """Initialize the low-level object references for this relation.

        This function is used only internally and called during the forward initialization.
        """
        if self.schema is None:
            msg = 'The target schema of the relation is not bound to a database'
            raise RelationError(msg)

        try:
            db = self.schema.get_db()
        except SchemaNotBoundError as e:
            msg = f"A database with schema '{self.schema.__name__}' needs to be created first!"
            raise RelationError(msg) from e

        if self.two_way_prop:
            self.obj_ref = obj_schema.DualPropertyRelation.build(db.id)
        else:
            self.obj_ref = obj_schema.SinglePropertyRelation.build(db.id)

    def _update_bwd_rel(self) -> None:
        """Change the default name of a two-way relation target to the defined one."""
        if self.prop_ref is None:
            msg = 'Trying to inialize a backward relation for one-way relation that is not bound to a property'
            raise SchemaError(msg)

        if not isinstance(self.obj_ref.relation, obj_schema.DualPropertyRelation):
            msg = f'Trying to inialize backward relation for one-way relation {self.prop_ref.name}'
            raise SchemaError(msg)

        obj_synced_property_name = self.obj_ref.relation.dual_property.synced_property_name
        two_way_prop_name = self._two_way_prop.name if self._two_way_prop else None

        if obj_synced_property_name != two_way_prop_name:
            session = get_active_session()

            if self.schema is None:
                msg = 'The target schema of the relation is not bound to a database'
                raise RelationError(msg)

            # change the old default name in the target schema to what was passed during initialization
            other_db = self.schema.get_db()

            if (prop_id := self.obj_ref.relation.dual_property.synced_property_id) is None:
                msg = 'No synced property ID found in the relation object'
                raise SchemaError(msg)

            schema_dct = {prop_id: obj_schema.RenameProp(name=two_way_prop_name)}
            session.api.databases.update(db=other_db.obj_ref, schema=schema_dct)
            other_db.schema._set_obj_refs()

            our_db = self.prop_ref._schema.get_db()
            session.api.databases.update(db=our_db.obj_ref, schema={})  # sync obj_ref
            our_db.schema._set_obj_refs()


class RollupError(SchemaError):
    """Error if definition of rollup is wrong."""


class Rollup(PropertyType[obj_schema.Rollup], wraps=obj_schema.Rollup):
    """Defines the rollup property in a database.

    If the relation propery is a self-referencing relation, i.e. `uno.PropType.Relation(uno.SelfRef)` in the schema,
    then the `property` must be a `str` of the corresponding property name.
    """

    def __init__(
        self, relation_prop: Property, rollup_prop: Property, calculate: AggFunc | str = AggFunc.SHOW_ORIGINAL
    ):
        if not isinstance(relation_prop.type, Relation):
            msg = f'Relation `{relation_prop.name}` must be of type Relation'
            raise RollupError(msg)

        calculate = AggFunc.from_alias(calculate) if not isinstance(calculate, AggFunc) else calculate

        # ToDo: One could check here if property really is a property in the database where relation points to
        super().__init__(relation_prop.name, rollup_prop.name, calculate)

    @property
    def _is_init(self) -> bool:
        """Determines if the relation of the rollup is already initialized."""
        try:
            return self.relation_prop.type._is_init
        except SchemaError:  # DB is not bound yet
            return False

    @property
    def is_self_ref(self) -> bool:
        """Determines if this rollup is self-referencing the same schema."""
        return cast(Relation, self.relation_prop.type).is_self_ref

    @property
    def relation_prop(self) -> Property:
        """Return the relation property object of the rollup."""
        if self.prop_ref is not None:
            return self.prop_ref._schema.get_prop(self.obj_ref.rollup.relation_property_name)
        else:
            msg = 'Rollup property not bound to a `Property` object'
            raise RollupError(msg)

    @property
    def rollup_prop(self) -> Property:
        """Return the rollup property object of the rollup."""
        if self.prop_ref is not None:
            return self.prop_ref._schema.get_prop(self.obj_ref.rollup.rollup_property_name)
        else:
            msg = 'Rollup property not bound to a `Property` object'
            raise RollupError(msg)

    @property
    def calculate(self) -> AggFunc:
        return self.obj_ref.rollup.function


class CreatedTime(PropertyType[obj_schema.CreatedTime], wraps=obj_schema.CreatedTime):
    """Defines the created-time property in a database."""


class CreatedBy(PropertyType[obj_schema.CreatedBy], wraps=obj_schema.CreatedBy):
    """Defines the created-by property in a database."""


class LastEditedBy(PropertyType[obj_schema.LastEditedBy], wraps=obj_schema.LastEditedBy):
    """Defines the last-edited-by property in a database."""


class LastEditedTime(PropertyType[obj_schema.LastEditedTime], wraps=obj_schema.LastEditedTime):
    """Defines the last-edited-time property in a database."""


class ID(PropertyType[obj_schema.UniqueID], wraps=obj_schema.UniqueID):
    """Defines a unique ID property in a database."""

    allowed_at_creation = False

    @property
    def prefix(self) -> str:
        opt_prefix = self.obj_ref.unique_id.prefix
        return '' if opt_prefix is None else opt_prefix


class Verification(PropertyType[obj_schema.Verification], wraps=obj_schema.Verification):
    """Defines a unique ID property in a database."""

    allowed_at_creation = False


class Button(PropertyType[obj_schema.Button], wraps=obj_schema.Button):
    """Defines a button property in a database."""

    allowed_at_creation = False


class DefaultSchema(Schema, db_title=None):
    """Default database schema of Notion.

    As inferred by just creating an empty database in the Notion UI.
    """

    name = Property('Name', Title())
    tags = Property('Tags', MultiSelect([]))


class PropType:
    """Namespace class of all property types of a database for easier access."""

    Title = Title
    Text = Text
    Number = Number
    Select = Select
    MultiSelect = MultiSelect
    Status = Status
    Date = Date
    Person = Person
    Files = Files
    Checkbox = Checkbox
    Email = Email
    URL = URL
    Phone = Phone
    Formula = Formula
    Relation = Relation
    Rollup = Rollup
    CreatedTime = CreatedTime
    CreatedBy = CreatedBy
    LastEditedTime = LastEditedTime
    LastEditedBy = LastEditedBy
    ID = ID
    Verification = Verification



"""Functionality to initialize a Notion Session."""

from __future__ import annotations

import logging
import time
from threading import RLock
from types import TracebackType
from typing import Any, ClassVar, cast
from uuid import UUID

import httpx
import notion_client
from notion_client.errors import APIResponseError

from ultimate_notion.blocks import Block, DataObject
from ultimate_notion.config import Config, activate_debug_mode, get_or_create_cfg
from ultimate_notion.database import Database
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import create_notion_client
from ultimate_notion.obj_api import query as obj_query
from ultimate_notion.obj_api.endpoints import NotionAPI
from ultimate_notion.obj_api.objects import UnknownUser as UnknownUserObj
from ultimate_notion.obj_api.objects import get_uuid
from ultimate_notion.page import Page
from ultimate_notion.props import Title
from ultimate_notion.rich_text import Text
from ultimate_notion.schema import DefaultSchema, Schema
from ultimate_notion.user import User
from ultimate_notion.utils import SList

_logger = logging.getLogger(__name__)


class SessionError(Exception):
    """Raised when there are issues with the Notion session."""


class Session:
    """A session for the Notion API.

    The session keeps tracks of all objects, e.g. pages, databases, etc.
    in an object store to avoid unnecessary calls to the API.
    """

    client: notion_client.Client
    api: NotionAPI
    _active_session: Session | None = None
    _lock = RLock()
    _own_bot_id: UUID | None = None
    cache: ClassVar[dict[UUID, DataObject | User]] = {}

    def __init__(self, cfg: Config | None = None, *, client: notion_client.Client | None = None, **kwargs: Any):
        """Initialize the `Session` object and the raw `api` endpoints.

        Args:
            cfg: configuration object
            **kwargs: Arguments for the [Notion SDK Client](https://ramnes.github.io/notion-sdk-py/reference/client/)
        """
        cfg = get_or_create_cfg() if cfg is None else cfg

        Session._initialize_once(self)

        if cfg.ultimate_notion.debug:
            activate_debug_mode()

        self.client = create_notion_client(cfg, **kwargs) if client is None else client
        self.api = NotionAPI(self.client)

    @classmethod
    def _initialize_once(cls, instance: Session):
        _logger.info('Initializing Notion session.')
        with Session._lock:
            if Session._active_session and Session._active_session is not instance:
                msg = 'Cannot initialize multiple Sessions at once'
                raise ValueError(msg)
            else:
                Session._active_session = instance

    @classmethod
    def get_active(cls) -> Session:
        """Return the current active session or None."""
        with Session._lock:
            if Session._active_session:
                return Session._active_session
            else:
                msg = 'There is no activate Session'
                raise ValueError(msg)

    @classmethod
    def get_or_create(cls, *args, **kwargs) -> Session:
        """Return the current active session or create a new session."""
        with Session._lock:
            if Session._active_session:
                return Session._active_session
            else:
                return cls(*args, **kwargs)

    def __enter__(self) -> Session:
        # Do nothing here as `__init__` created a client already and calling
        # `self.client.__enter__()` would initialize another client internally
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        self.close()

    def close(self):
        """Close the session and release resources."""
        _logger.info('Closing connection to Notion.')
        self.client.close()
        Session._active_session = None
        Session.cache.clear()
        Session._own_bot_id = None

    def is_closed(self) -> bool:
        """Determine if the session is closed or not."""
        try:
            self.raise_for_status()
        except RuntimeError:  # used by httpx client in case of closed connection
            return True
        else:
            return False

    def raise_for_status(self):
        """Confirm that the session is active and raise otherwise.

        Raises SessionError if there is a problem, otherwise returns None.
        """
        _logger.info('Checking connection to Notion.')
        try:
            self.whoami()
        except httpx.ConnectError as err:
            msg = 'Unable to connect to Notion'
            raise SessionError(msg) from err
        except APIResponseError as err:
            msg = 'Invalid API reponse'
            raise SessionError(msg) from err

    def get_block(self, block_ref: UUID | str, *, use_cache: bool = True) -> Block:
        """Retrieve a single block by an object reference."""
        block_uuid = get_uuid(block_ref)
        if use_cache and block_uuid in self.cache:
            _logger.info(f'Retrieving cached block with id `{block_uuid}`.')
            block = cast(Block, self.cache[block_uuid])
        else:
            _logger.info(f'Retrieving block with id `{block_uuid}`.')
            block = Block.wrap_obj_ref(self.api.blocks.retrieve(block_uuid))
            self.cache[block.id] = block
        _logger.info(f'Retrieved `{type(block)}` block.')
        return block

    # ToDo: Provide a title and description for the database that overwrites the values from the schema
    def create_db(self, parent: Page, schema: type[Schema] | None = None) -> Database:
        """Create a new database within a page."""
        # Implementation:
        # 1. initialize external forward relations, i.e. relations pointing to other databases
        # 2. create the database using a Notion API call and potential external forward relations
        # 3. initialize self-referencing forward relations
        # 4. create properties with self-referencing forward relations using an update call
        # 5. update the backward references, i.e. two-way relations, using an update call
        if schema:
            _logger.info(f'Creating database `{schema.db_title}` in page `{parent.title}` with schema:\n{schema}')
            schema._init_fwd_rels()
            schema_dct = {prop.name: prop.type.obj_ref for prop in schema._get_init_props()}
            title = schema.db_title.obj_ref if schema.db_title is not None else None
            db_obj = self.api.databases.create(parent=parent.obj_ref, title=title, schema=schema_dct)
            if schema.db_desc:
                db_obj = self.api.databases.update(db_obj, description=schema.db_desc.obj_ref)
        else:
            _logger.info(f'Creating database in page `{parent.title}` with default schema.')
            schema_dct = {prop.name: prop.type.obj_ref for prop in DefaultSchema._get_init_props()}
            db_obj = self.api.databases.create(parent=parent.obj_ref, schema=schema_dct)

        db: Database = Database.wrap_obj_ref(db_obj)

        if schema:
            db._set_schema(schema, during_init=True)  # schema is thus bound to the database
            schema._init_self_refs()
            schema._init_self_ref_rollups()
            schema._update_bwd_rels()

        self.cache[db.id] = db
        return db

    def create_dbs(self, parents: Page | list[Page], schemas: list[type[Schema]]) -> list[Database]:
        """Create new databases in the right order in case there a relations between them."""
        # ToDo: Implement
        raise NotImplementedError

    def search_db(
        self, db_name: str | None = None, *, exact: bool = True, reverse: bool = False, deleted: bool = False
    ) -> SList[Database]:
        """Search a database by name or return all if `db_name` is None.

        Args:
            db_name: name/title of the database, return all if `None`
            exact: perform an exact search, not only a substring match
            reverse: search in the reverse order, i.e. the least recently edited results first
            deleted: include deleted databases in search
        """
        _logger.info(f'Searching for database with name `{db_name}`.')
        query: obj_query.SearchQueryBuilder[obj_blocks.Database] = self.api.search(db_name).filter(db_only=True)
        if reverse:
            query.sort(ascending=True)
        dbs = [cast(Database, self.cache.setdefault(db.id, Database.wrap_obj_ref(db))) for db in query.execute()]
        if exact and db_name is not None:
            dbs = [db for db in dbs if db.title == db_name]
        if not deleted:
            dbs = [db for db in dbs if not db.is_deleted]
        return SList(dbs)

    def get_db(self, db_ref: UUID | str, *, use_cache: bool = True) -> Database:
        """Retrieve Notion database by uuid"""
        db_uuid = get_uuid(db_ref)
        if use_cache and db_uuid in self.cache:
            _logger.info(f'Retrieving cached database with id `{db_uuid}`.')
            db = cast(Database, self.cache[db_uuid])
        else:
            _logger.info(f'Retrieving database with id `{db_uuid}`.')
            db = Database.wrap_obj_ref(self.api.databases.retrieve(db_uuid))
            self.cache[db.id] = db
        _logger.info(f'Retrieved database `{db.title}`.')
        return db

    def get_or_create_db(self, parent: Page, schema: type[Schema]) -> Database:
        """Get or create the database."""
        dbs = SList(db for db in self.search_db(schema.db_title) if db.parent == parent)
        if len(dbs) == 0:
            db = self.create_db(parent, schema)
            while not [db for db in self.search_db(schema.db_title) if db.parent == parent]:
                _logger.info(f'Waiting for database `{db.title}` to be fully created.')
                time.sleep(1)
            return db
        else:
            db = dbs.item()
            db.schema = schema
            return db

    def search_page(self, title: str | None = None, *, exact: bool = True, reverse: bool = False) -> SList[Page]:
        """Search a page by name. Deleted pages, i.e. in trash, are not included in the search.

        Args:
            title: title of the page, return all if `None`
            exact: perform an exact search, not only a substring match
            reverse: search in the reverse order, i.e. the least recently edited results first
        """
        _logger.info(f'Searching for page with title `{title}`.')
        query: obj_query.SearchQueryBuilder[obj_blocks.Page] = self.api.search(title).filter(page_only=True)
        if reverse:
            query.sort(ascending=True)
        pages = [
            cast(Page, self.cache.setdefault(page_obj.id, Page.wrap_obj_ref(page_obj))) for page_obj in query.execute()
        ]
        if exact and title is not None:
            pages = [page for page in pages if page.title == title]
        return SList(pages)

    def get_page(self, page_ref: UUID | str, *, use_cache: bool = True) -> Page:
        """Retrieve a page by uuid."""
        page_uuid = get_uuid(page_ref)
        if use_cache and page_uuid in self.cache:
            _logger.info(f'Retrieving cached page with id `{page_uuid}`.')
            page = cast(Page, self.cache[page_uuid])
        else:
            _logger.info(f'Retrieving page with id `{page_uuid}`.')
            page = Page.wrap_obj_ref(self.api.pages.retrieve(page_uuid))
            self.cache[page.id] = page
        _logger.info(f'Retrieved page `{page.title}`.')
        return page

    def create_page(self, parent: Page | Database, title: Text | str | None = None) -> Page:
        """Create a new page in a parent page or database."""
        _logger.info(f'Creating page with title `{title}` in parent `{parent.title}`.')
        title_obj = title if title is None else Title(title).obj_ref
        page = Page.wrap_obj_ref(self.api.pages.create(parent=parent.obj_ref, title=title_obj))
        self.cache[page.id] = page
        return page

    def get_user(self, user_ref: UUID | str, *, use_cache: bool = True) -> User:
        """Get a user by uuid.

        !!! warning

            Trying to retrieve yourself, i.e. the bot integration, only works if `use_cache` is true,
            since the low-level api, i.e. `api.users.retrieve()` does not work for the bot integration.
            Better use `whoami()` to get the bot integration user object.
        """
        user_uuid = get_uuid(user_ref)
        self.whoami()  # make sure cache is filled with the uuid of the bot integration

        if use_cache and user_uuid in self.cache:
            _logger.info(f'Retrieving cached user with id `{user_uuid}`.')
            user = cast(User, self.cache[user_uuid])
        else:
            _logger.info(f'Retrieving user with id `{user_uuid}`.')
            try:
                user_obj = self.api.users.retrieve(user_uuid)
            except APIResponseError:
                _logger.warning(f'User with id {user_uuid} not found!')
                user_obj = UnknownUserObj(id=user_uuid, object='user', type='unknown')
            user = User.wrap_obj_ref(user_obj)
            self.cache[user.id] = user

        _logger.info(f'Retrieved user `{user}`.')
        return user

    def search_user(self, name: str) -> SList[User]:
        """Search a user by name."""
        _logger.info(f'Searching for user with name `{name}`.')
        return SList(user for user in self.all_users() if user.name == name)

    def all_users(self) -> list[User]:
        """Retrieve all users of this workspace."""
        _logger.info('Retrieving all users.')
        return [cast(User, self.cache.setdefault(user.id, User.wrap_obj_ref(user))) for user in self.api.users.list()]

    def whoami(self) -> User:
        """Return the user object of this bot."""
        _logger.info('Retrieving user of this integration.')
        if self._own_bot_id is None:
            user = self.api.users.me()
            self._own_bot_id = user.id
            return cast(User, self.cache.setdefault(user.id, User.wrap_obj_ref(user)))
        else:
            return cast(User, self.cache[self._own_bot_id])
